"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[9666],{23001:function(e,n,t){t.d(n,{X:function(){return i}});var r=t(67294);function i(){let e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:768,[n,t]=(0,r.useState)(!1);return(0,r.useEffect)(()=>{let n=()=>{t(window.innerWidth<e)};return n(),window.addEventListener("resize",n),()=>{window.removeEventListener("resize",n)}},[e]),n}},32350:function(e,n,t){t.d(n,{cn:function(){return u}});var r=t(90512),i=t(98388);function u(){for(var e=arguments.length,n=Array(e),t=0;t<e;t++)n[t]=arguments[t];return(0,i.m6)((0,r.W)(n))}}}]);