"use strict";(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[8252],{88507:function(e,l,n){n.d(l,{Z:function(){return t}});/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let t=(0,n(60998).Z)("MessageSquare",[["path",{d:"M21 15a2 2 0 0 1-2 2H7l-4 4V5a2 2 0 0 1 2-2h14a2 2 0 0 1 2 2z",key:"1lielz"}]])},28586:function(e,l,n){n.d(l,{Z:function(){return t}});/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let t=(0,n(60998).Z)("Users",[["path",{d:"M16 21v-2a4 4 0 0 0-4-4H6a4 4 0 0 0-4 4v2",key:"1yyitq"}],["circle",{cx:"9",cy:"7",r:"4",key:"nufk8"}],["path",{d:"M22 21v-2a4 4 0 0 0-3-3.87",key:"kshegd"}],["path",{d:"M16 3.13a4 4 0 0 1 0 7.75",key:"1da9ce"}]])},43767:function(e,l,n){n.d(l,{Z:function(){return t}});/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let t=(0,n(60998).Z)("X",[["path",{d:"M18 6 6 18",key:"1bl5f8"}],["path",{d:"m6 6 12 12",key:"d8bk6v"}]])},12003:function(e,l,n){n.d(l,{j:function(){return a}});var t=n(90512);let u=e=>"boolean"==typeof e?`${e}`:0===e?"0":e,r=t.W,a=(e,l)=>n=>{var t;if((null==l?void 0:l.variants)==null)return r(e,null==n?void 0:n.class,null==n?void 0:n.className);let{variants:a,defaultVariants:i}=l,c=Object.keys(a).map(e=>{let l=null==n?void 0:n[e],t=null==i?void 0:i[e];if(null===l)return null;let r=u(l)||u(t);return a[e][r]}),s=n&&Object.entries(n).reduce((e,l)=>{let[n,t]=l;return void 0===t||(e[n]=t),e},{});return r(e,c,null==l?void 0:null===(t=l.compoundVariants)||void 0===t?void 0:t.reduce((e,l)=>{let{class:n,className:t,...u}=l;return Object.entries(u).every(e=>{let[l,n]=e;return Array.isArray(n)?n.includes({...i,...s}[l]):({...i,...s})[l]===n})?[...e,n,t]:e},[]),null==n?void 0:n.class,null==n?void 0:n.className)}}}]);