(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[2888],{83454:function(e,t,n){"use strict";var o,r;e.exports=(null==(o=n.g.process)?void 0:o.env)&&"object"==typeof(null==(r=n.g.process)?void 0:r.env)?n.g.process:n(77663)},91118:function(e,t,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/_app",function(){return n(32586)}])},19606:function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0}),function(e,t){for(var n in t)Object.defineProperty(e,n,{enumerable:!0,get:t[n]})}(t,{default:function(){return s},noSSR:function(){return l}});let o=n(10260);n(85893),n(67294);let r=o._(n(6119));function i(e){return{default:(null==e?void 0:e.default)||e}}function l(e,t){return delete t.webpack,delete t.modules,e(t)}function s(e,t){let n=r.default,o={loading:e=>{let{error:t,isLoading:n,pastDelay:o}=e;return null}};e instanceof Promise?o.loader=()=>e:"function"==typeof e?o.loader=e:"object"==typeof e&&(o={...o,...e});let s=(o={...o,...t}).loader;return(o.loadableGenerated&&(o={...o,...o.loadableGenerated},delete o.loadableGenerated),"boolean"!=typeof o.ssr||o.ssr)?n({...o,loader:()=>null!=s?s().then(i):Promise.resolve(i(()=>null))}):(delete o.webpack,delete o.modules,l(n,o))}("function"==typeof t.default||"object"==typeof t.default&&null!==t.default)&&void 0===t.default.__esModule&&(Object.defineProperty(t.default,"__esModule",{value:!0}),Object.assign(t.default,t),e.exports=t.default)},16725:function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"LoadableContext",{enumerable:!0,get:function(){return o}});let o=n(10260)._(n(67294)).default.createContext(null)},6119:function(e,t,n){"use strict";/**
@copyright (c) 2017-present James Kyle <me@thejameskyle.com>
 MIT License
 Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:
 The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
*/Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"default",{enumerable:!0,get:function(){return h}});let o=n(10260)._(n(67294)),r=n(16725),i=[],l=[],s=!1;function a(e){let t=e(),n={loading:!0,loaded:null,error:null};return n.promise=t.then(e=>(n.loading=!1,n.loaded=e,e)).catch(e=>{throw n.loading=!1,n.error=e,e}),n}class c{promise(){return this._res.promise}retry(){this._clearTimeouts(),this._res=this._loadFn(this._opts.loader),this._state={pastDelay:!1,timedOut:!1};let{_res:e,_opts:t}=this;e.loading&&("number"==typeof t.delay&&(0===t.delay?this._state.pastDelay=!0:this._delay=setTimeout(()=>{this._update({pastDelay:!0})},t.delay)),"number"==typeof t.timeout&&(this._timeout=setTimeout(()=>{this._update({timedOut:!0})},t.timeout))),this._res.promise.then(()=>{this._update({}),this._clearTimeouts()}).catch(e=>{this._update({}),this._clearTimeouts()}),this._update({})}_update(e){this._state={...this._state,error:this._res.error,loaded:this._res.loaded,loading:this._res.loading,...e},this._callbacks.forEach(e=>e())}_clearTimeouts(){clearTimeout(this._delay),clearTimeout(this._timeout)}getCurrentValue(){return this._state}subscribe(e){return this._callbacks.add(e),()=>{this._callbacks.delete(e)}}constructor(e,t){this._loadFn=e,this._opts=t,this._callbacks=new Set,this._delay=null,this._timeout=null,this.retry()}}function u(e){return function(e,t){let n=Object.assign({loader:null,loading:null,delay:200,timeout:null,webpack:null,modules:null},t),i=null;function a(){if(!i){let t=new c(e,n);i={getCurrentValue:t.getCurrentValue.bind(t),subscribe:t.subscribe.bind(t),retry:t.retry.bind(t),promise:t.promise.bind(t)}}return i.promise()}if(!s){let e=n.webpack?n.webpack():n.modules;e&&l.push(t=>{for(let n of e)if(t.includes(n))return a()})}function u(e,t){!function(){a();let e=o.default.useContext(r.LoadableContext);e&&Array.isArray(n.modules)&&n.modules.forEach(t=>{e(t)})}();let l=o.default.useSyncExternalStore(i.subscribe,i.getCurrentValue,i.getCurrentValue);return o.default.useImperativeHandle(t,()=>({retry:i.retry}),[]),o.default.useMemo(()=>{var t;return l.loading||l.error?o.default.createElement(n.loading,{isLoading:l.loading,pastDelay:l.pastDelay,timedOut:l.timedOut,error:l.error,retry:i.retry}):l.loaded?o.default.createElement((t=l.loaded)&&t.default?t.default:t,e):null},[e,l])}return u.preload=()=>a(),u.displayName="LoadableComponent",o.default.forwardRef(u)}(a,e)}function d(e,t){let n=[];for(;e.length;){let o=e.pop();n.push(o(t))}return Promise.all(n).then(()=>{if(e.length)return d(e,t)})}u.preloadAll=()=>new Promise((e,t)=>{d(i).then(e,t)}),u.preloadReady=e=>(void 0===e&&(e=[]),new Promise(t=>{let n=()=>(s=!0,t());d(l,e).then(n,n)})),window.__NEXT_PRELOADREADY=u.preloadReady;let h=u},93225:function(e,t,n){"use strict";n.d(t,{$m:function(){return a},Bo:function(){return l},GW:function(){return r},Z2:function(){return c},f4:function(){return o},iW:function(){return s},mF:function(){return i}}),n(83454);let o=function(){{let e=window.location.pathname,t=e.indexOf("/dashboard");if(-1!==t){let n=e.substring(0,t);return"".concat(n,"/internal/dashboard")}return"/internal/dashboard"}}(),r=function(){{let e=window.location.pathname,t=e.indexOf("/dashboard");return -1!==t?e.substring(0,t+10):"/dashboard"}}();"/api/v1".replace(/^http/,"ws");let i="ClusterDoesNotExist",l="NotSupportedError",s="ClusterNotUpError",a=["AWS","Azure","GCP","IBM","Lambda","SCP","OCI","RunPod","VAST","vSphere","Cudo","FluidStack","Paperspace","DO","Nebius"],c=Object.fromEntries([...a.map(e=>[e.toLowerCase(),e]),["kubernetes","Kubernetes"],["ssh","SSH Node Pool"]])},6556:function(e,t,n){"use strict";n.d(t,{n:function(){return l}});var o=n(67294);let r="skypilot-dashboard-first-visit",i="skypilot-dashboard-tour-completed";function l(){let[e,t]=(0,o.useState)(!1),[n,l]=(0,o.useState)(!1),[s,a]=(0,o.useState)(!1);return(0,o.useEffect)(()=>{a(!0);let e=sessionStorage.getItem(r),n=localStorage.getItem(i);e||(t(!0),sessionStorage.setItem(r,"true")),n&&l(!0)},[]),{isFirstVisit:!!s&&e,tourCompleted:!!s&&n,shouldShowTourPrompt:s&&e&&!n,shouldPulseHelpButton:s&&e&&!n,markTourCompleted:()=>{localStorage.setItem(i,"true"),l(!0)},resetFirstVisit:()=>{sessionStorage.removeItem(r),localStorage.removeItem(i),t(!0),l(!1)},isClient:s}}},94849:function(e,t,n){"use strict";let o,r;n.d(t,{C:function(){return t5},r:function(){return t3}});var i,l,s=n(85893),a=n(67294),c=n(11163);function u(e){return e instanceof HTMLElement}function d(e){return"function"==typeof e}function h(e){return"string"==typeof e}function f(e){return void 0===e}class p{on(e,t,n,o=!1){var r;return f(this.bindings)&&(this.bindings={}),f(this.bindings[e])&&(this.bindings[e]=[]),null==(r=this.bindings[e])||r.push({handler:t,ctx:n,once:o}),this}once(e,t,n){return this.on(e,t,n,!0)}off(e,t){if(f(this.bindings)||f(this.bindings[e]))return this;if(f(t))delete this.bindings[e];else{var n;null==(n=this.bindings[e])||n.forEach((n,o)=>{if(n.handler===t){var r;null==(r=this.bindings[e])||r.splice(o,1)}})}return this}trigger(e,...t){if(!f(this.bindings)&&this.bindings[e]){var n;null==(n=this.bindings[e])||n.forEach((n,o)=>{let{ctx:r,handler:i,once:l}=n;if(i.apply(r||this,t),l){var s;null==(s=this.bindings[e])||s.splice(o,1)}})}return this}}function m(){return(m=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var o in n)({}).hasOwnProperty.call(n,o)&&(e[o]=n[o])}return e}).apply(null,arguments)}function g(e,t){if(null==e)return{};var n={};for(var o in e)if(({}).hasOwnProperty.call(e,o)){if(t.includes(o))continue;n[o]=e[o]}return n}let y={defaultMerge:Symbol("deepmerge-ts: default merge"),skip:Symbol("deepmerge-ts: skip")};function b(e,t){return t}function v(e,t){return e.filter(e=>void 0!==e)}function x(e){return"object"!=typeof e||null===e?0:Array.isArray(e)?2:!function(e){if(!E.has(Object.prototype.toString.call(e)))return!1;let{constructor:t}=e;if(void 0===t)return!0;let n=t.prototype;return!!(null!==n&&"object"==typeof n&&E.has(Object.prototype.toString.call(n))&&n.hasOwnProperty("isPrototypeOf"))}(e)?e instanceof Set?3:e instanceof Map?4:5:1}function w(e){return{*[Symbol.iterator](){for(let t of e)for(let e of t)yield e}}}y.defaultMerge,(i=l||(l={}))[i.NOT=0]="NOT",i[i.RECORD=1]="RECORD",i[i.ARRAY=2]="ARRAY",i[i.SET=3]="SET",i[i.MAP=4]="MAP",i[i.OTHER=5]="OTHER";let E=new Set(["[object Object]","[object Module]"]),S={mergeRecords:function(e,t,n){let o={};for(let r of function(e){let t=new Set;for(let n of e)for(let e of[...Object.keys(n),...Object.getOwnPropertySymbols(n)])t.add(e);return t}(e)){let i=[];for(let t of e)"object"==typeof t&&Object.prototype.propertyIsEnumerable.call(t,r)&&i.push(t[r]);if(0===i.length)continue;let l=t.metaDataUpdater(n,{key:r,parents:e}),s=$(i,t,l);s!==y.skip&&("__proto__"===r?Object.defineProperty(o,r,{value:s,configurable:!0,enumerable:!0,writable:!0}):o[r]=s)}return o},mergeArrays:function(e){return e.flat()},mergeSets:function(e){return new Set(w(e))},mergeMaps:function(e){return new Map(w(e))},mergeOthers:function(e){return e.at(-1)}};function T(...e){return(function(e,t){var n,o,r;let i={defaultMergeFunctions:S,mergeFunctions:m({},S,Object.fromEntries(Object.entries(e).filter(([e,t])=>Object.hasOwn(S,e)).map(([e,t])=>!1===t?[e,S.mergeOthers]:[e,t]))),metaDataUpdater:null!=(n=e.metaDataUpdater)?n:b,deepmerge:l,useImplicitDefaultMerging:null!=(o=e.enableImplicitDefaultMerging)&&o,filterValues:!1===e.filterValues?void 0:null!=(r=e.filterValues)?r:v,actions:y};function l(...e){return $(e,i,void 0)}return l})({})(...e)}function $(e,t,n){var o;let r=null!=(o=null==t.filterValues?void 0:t.filterValues(e,n))?o:e;if(0===r.length)return;if(1===r.length)return k(r,t,n);let i=x(r[0]);if(0!==i&&5!==i){for(let e=1;e<r.length;e++)if(x(r[e])!==i)return k(r,t,n)}switch(i){case 1:return function(e,t,n){let o=t.mergeFunctions.mergeRecords(e,t,n);return o===y.defaultMerge||t.useImplicitDefaultMerging&&void 0===o&&t.mergeFunctions.mergeRecords!==t.defaultMergeFunctions.mergeRecords?t.defaultMergeFunctions.mergeRecords(e,t,n):o}(r,t,n);case 2:return function(e,t,n){let o=t.mergeFunctions.mergeArrays(e,t,n);return o===y.defaultMerge||t.useImplicitDefaultMerging&&void 0===o&&t.mergeFunctions.mergeArrays!==t.defaultMergeFunctions.mergeArrays?t.defaultMergeFunctions.mergeArrays(e):o}(r,t,n);case 3:return function(e,t,n){let o=t.mergeFunctions.mergeSets(e,t,n);return o===y.defaultMerge||t.useImplicitDefaultMerging&&void 0===o&&t.mergeFunctions.mergeSets!==t.defaultMergeFunctions.mergeSets?t.defaultMergeFunctions.mergeSets(e):o}(r,t,n);case 4:return function(e,t,n){let o=t.mergeFunctions.mergeMaps(e,t,n);return o===y.defaultMerge||t.useImplicitDefaultMerging&&void 0===o&&t.mergeFunctions.mergeMaps!==t.defaultMergeFunctions.mergeMaps?t.defaultMergeFunctions.mergeMaps(e):o}(r,t,n);default:return k(r,t,n)}}function k(e,t,n){let o=t.mergeFunctions.mergeOthers(e,t,n);return o===y.defaultMerge||t.useImplicitDefaultMerging&&void 0===o&&t.mergeFunctions.mergeOthers!==t.defaultMergeFunctions.mergeOthers?t.defaultMergeFunctions.mergeOthers(e):o}function _(e){let t=Object.getOwnPropertyNames(e.constructor.prototype);for(let n=0;n<t.length;n++){let o=t[n],r=e[o];"constructor"!==o&&"function"==typeof r&&(e[o]=r.bind(e))}return e}class C{constructor(e){}}class O{constructor(e,t){}}function P(e){return h(e)&&""!==e?"-"!==e.charAt(e.length-1)?`${e}-`:e:""}function A(e){return null==e||!e.element||!e.on}function L(){let e=Date.now();return"xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g,t=>{let n=(e+16*Math.random())%16|0;return e=Math.floor(e/16),("x"==t?n:3&n|8).toString(16)})}let M=["top","right","bottom","left"].reduce((e,t)=>e.concat(t,t+"-start",t+"-end"),[]),I=Math.min,R=Math.max,q=Math.round,B=Math.floor,H=e=>({x:e,y:e}),j={left:"right",right:"left",bottom:"top",top:"bottom"},D={start:"end",end:"start"};function F(e,t){return"function"==typeof e?e(t):e}function N(e){return e.split("-")[0]}function V(e){return e.split("-")[1]}function W(e){return"x"===e?"y":"x"}function z(e){return"y"===e?"height":"width"}function U(e){return["top","bottom"].includes(N(e))?"y":"x"}function Y(e,t,n){void 0===n&&(n=!1);let o=V(e),r=W(U(e)),i=z(r),l="x"===r?o===(n?"end":"start")?"right":"left":"start"===o?"bottom":"top";return t.reference[i]>t.floating[i]&&(l=K(l)),[l,K(l)]}function G(e){return e.replace(/start|end/g,e=>D[e])}function K(e){return e.replace(/left|right|bottom|top/g,e=>j[e])}function J(e){return"number"!=typeof e?m({top:0,right:0,bottom:0,left:0},e):{top:e,right:e,bottom:e,left:e}}function X(e){let{x:t,y:n,width:o,height:r}=e;return{width:o,height:r,top:n,left:t,right:t+o,bottom:n+r,x:t,y:n}}let Z=["crossAxis","alignment","allowedPlacements","autoAlignment"],Q=["mainAxis","crossAxis","fallbackPlacements","fallbackStrategy","fallbackAxisSideDirection","flipAlignment"],ee=["mainAxis","crossAxis","limiter"];function et(e,t,n){let o,{reference:r,floating:i}=e,l=U(t),s=W(U(t)),a=z(s),c=N(t),u="y"===l,d=r.x+r.width/2-i.width/2,h=r.y+r.height/2-i.height/2,f=r[a]/2-i[a]/2;switch(c){case"top":o={x:d,y:r.y-i.height};break;case"bottom":o={x:d,y:r.y+r.height};break;case"right":o={x:r.x+r.width,y:h};break;case"left":o={x:r.x-i.width,y:h};break;default:o={x:r.x,y:r.y}}switch(V(t)){case"start":o[s]-=f*(n&&u?-1:1);break;case"end":o[s]+=f*(n&&u?-1:1)}return o}let en=async(e,t,n)=>{let{placement:o="bottom",strategy:r="absolute",middleware:i=[],platform:l}=n,s=i.filter(Boolean),a=await (null==l.isRTL?void 0:l.isRTL(t)),c=await l.getElementRects({reference:e,floating:t,strategy:r}),{x:u,y:d}=et(c,o,a),h=o,f={},p=0;for(let n=0;n<s.length;n++){let{name:i,fn:g}=s[n],{x:y,y:b,data:v,reset:x}=await g({x:u,y:d,initialPlacement:o,placement:h,strategy:r,middlewareData:f,rects:c,platform:l,elements:{reference:e,floating:t}});u=null!=y?y:u,d=null!=b?b:d,f=m({},f,{[i]:m({},f[i],v)}),x&&p<=50&&(p++,"object"==typeof x&&(x.placement&&(h=x.placement),x.rects&&(c=!0===x.rects?await l.getElementRects({reference:e,floating:t,strategy:r}):x.rects),{x:u,y:d}=et(c,h,a)),n=-1)}return{x:u,y:d,placement:h,strategy:r,middlewareData:f}};async function eo(e,t){var n;void 0===t&&(t={});let{x:o,y:r,platform:i,rects:l,elements:s,strategy:a}=e,{boundary:c="clippingAncestors",rootBoundary:u="viewport",elementContext:d="floating",altBoundary:h=!1,padding:f=0}=F(t,e),p=J(f),m=s[h?"floating"===d?"reference":"floating":d],g=X(await i.getClippingRect({element:null==(n=await (null==i.isElement?void 0:i.isElement(m)))||n?m:m.contextElement||await (null==i.getDocumentElement?void 0:i.getDocumentElement(s.floating)),boundary:c,rootBoundary:u,strategy:a})),y="floating"===d?{x:o,y:r,width:l.floating.width,height:l.floating.height}:l.reference,b=await (null==i.getOffsetParent?void 0:i.getOffsetParent(s.floating)),v=await (null==i.isElement?void 0:i.isElement(b))&&await (null==i.getScale?void 0:i.getScale(b))||{x:1,y:1},x=X(i.convertOffsetParentRelativeRectToViewportRelativeRect?await i.convertOffsetParentRelativeRectToViewportRelativeRect({elements:s,rect:y,offsetParent:b,strategy:a}):y);return{top:(g.top-x.top+p.top)/v.y,bottom:(x.bottom-g.bottom+p.bottom)/v.y,left:(g.left-x.left+p.left)/v.x,right:(x.right-g.right+p.right)/v.x}}function er(){return"undefined"!=typeof window}function ei(e){return ea(e)?(e.nodeName||"").toLowerCase():"#document"}function el(e){var t;return(null==e||null==(t=e.ownerDocument)?void 0:t.defaultView)||window}function es(e){var t;return null==(t=(ea(e)?e.ownerDocument:e.document)||window.document)?void 0:t.documentElement}function ea(e){return!!er()&&(e instanceof Node||e instanceof el(e).Node)}function ec(e){return!!er()&&(e instanceof Element||e instanceof el(e).Element)}function eu(e){return!!er()&&(e instanceof HTMLElement||e instanceof el(e).HTMLElement)}function ed(e){return!!er()&&"undefined"!=typeof ShadowRoot&&(e instanceof ShadowRoot||e instanceof el(e).ShadowRoot)}function eh(e){let{overflow:t,overflowX:n,overflowY:o,display:r}=ey(e);return/auto|scroll|overlay|hidden|clip/.test(t+o+n)&&!["inline","contents"].includes(r)}function ef(e){return[":popover-open",":modal"].some(t=>{try{return e.matches(t)}catch(e){return!1}})}function ep(e){let t=em(),n=ec(e)?ey(e):e;return"none"!==n.transform||"none"!==n.perspective||!!n.containerType&&"normal"!==n.containerType||!t&&!!n.backdropFilter&&"none"!==n.backdropFilter||!t&&!!n.filter&&"none"!==n.filter||["transform","perspective","filter"].some(e=>(n.willChange||"").includes(e))||["paint","layout","strict","content"].some(e=>(n.contain||"").includes(e))}function em(){return"undefined"!=typeof CSS&&!!CSS.supports&&CSS.supports("-webkit-backdrop-filter","none")}function eg(e){return["html","body","#document"].includes(ei(e))}function ey(e){return el(e).getComputedStyle(e)}function eb(e){return ec(e)?{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}:{scrollLeft:e.scrollX,scrollTop:e.scrollY}}function ev(e){if("html"===ei(e))return e;let t=e.assignedSlot||e.parentNode||ed(e)&&e.host||es(e);return ed(t)?t.host:t}function ex(e,t,n){var o;void 0===t&&(t=[]),void 0===n&&(n=!0);let r=function e(t){let n=ev(t);return eg(n)?t.ownerDocument?t.ownerDocument.body:t.body:eu(n)&&eh(n)?n:e(n)}(e),i=r===(null==(o=e.ownerDocument)?void 0:o.body),l=el(r);if(i){let e=ew(l);return t.concat(l,l.visualViewport||[],eh(r)?r:[],e&&n?ex(e):[])}return t.concat(r,ex(r,[],n))}function ew(e){return e.parent&&Object.getPrototypeOf(e.parent)?e.frameElement:null}function eE(e){let t=ey(e),n=parseFloat(t.width)||0,o=parseFloat(t.height)||0,r=eu(e),i=r?e.offsetWidth:n,l=r?e.offsetHeight:o,s=q(n)!==i||q(o)!==l;return s&&(n=i,o=l),{width:n,height:o,$:s}}function eS(e){return ec(e)?e:e.contextElement}function eT(e){let t=eS(e);if(!eu(t))return H(1);let n=t.getBoundingClientRect(),{width:o,height:r,$:i}=eE(t),l=(i?q(n.width):n.width)/o,s=(i?q(n.height):n.height)/r;return l&&Number.isFinite(l)||(l=1),s&&Number.isFinite(s)||(s=1),{x:l,y:s}}let e$=H(0);function ek(e){let t=el(e);return em()&&t.visualViewport?{x:t.visualViewport.offsetLeft,y:t.visualViewport.offsetTop}:e$}function e_(e,t,n,o){var r;void 0===t&&(t=!1),void 0===n&&(n=!1);let i=e.getBoundingClientRect(),l=eS(e),s=H(1);t&&(o?ec(o)&&(s=eT(o)):s=eT(e));let a=(void 0===(r=n)&&(r=!1),o&&(!r||o===el(l))&&r)?ek(l):H(0),c=(i.left+a.x)/s.x,u=(i.top+a.y)/s.y,d=i.width/s.x,h=i.height/s.y;if(l){let e=el(l),t=o&&ec(o)?el(o):o,n=e,r=ew(n);for(;r&&o&&t!==n;){let e=eT(r),t=r.getBoundingClientRect(),o=ey(r),i=t.left+(r.clientLeft+parseFloat(o.paddingLeft))*e.x,l=t.top+(r.clientTop+parseFloat(o.paddingTop))*e.y;c*=e.x,u*=e.y,d*=e.x,h*=e.y,c+=i,u+=l,r=ew(n=el(r))}}return X({width:d,height:h,x:c,y:u})}function eC(e,t){let n=eb(e).scrollLeft;return t?t.left+n:e_(es(e)).left+n}function eO(e,t,n){void 0===n&&(n=!1);let o=e.getBoundingClientRect();return{x:o.left+t.scrollLeft-(n?0:eC(e,o)),y:o.top+t.scrollTop}}function eP(e,t,n){let o;if("viewport"===t)o=function(e,t){let n=el(e),o=es(e),r=n.visualViewport,i=o.clientWidth,l=o.clientHeight,s=0,a=0;if(r){i=r.width,l=r.height;let e=em();(!e||e&&"fixed"===t)&&(s=r.offsetLeft,a=r.offsetTop)}return{width:i,height:l,x:s,y:a}}(e,n);else if("document"===t)o=function(e){let t=es(e),n=eb(e),o=e.ownerDocument.body,r=R(t.scrollWidth,t.clientWidth,o.scrollWidth,o.clientWidth),i=R(t.scrollHeight,t.clientHeight,o.scrollHeight,o.clientHeight),l=-n.scrollLeft+eC(e),s=-n.scrollTop;return"rtl"===ey(o).direction&&(l+=R(t.clientWidth,o.clientWidth)-r),{width:r,height:i,x:l,y:s}}(es(e));else if(ec(t))o=function(e,t){let n=e_(e,!0,"fixed"===t),o=n.top+e.clientTop,r=n.left+e.clientLeft,i=eu(e)?eT(e):H(1),l=e.clientWidth*i.x;return{width:l,height:e.clientHeight*i.y,x:r*i.x,y:o*i.y}}(t,n);else{let n=ek(e);o={x:t.x-n.x,y:t.y-n.y,width:t.width,height:t.height}}return X(o)}function eA(e){return"static"===ey(e).position}function eL(e,t){if(!eu(e)||"fixed"===ey(e).position)return null;if(t)return t(e);let n=e.offsetParent;return es(e)===n&&(n=n.ownerDocument.body),n}function eM(e,t){let n=el(e);if(ef(e))return n;if(!eu(e)){let t=ev(e);for(;t&&!eg(t);){if(ec(t)&&!eA(t))return t;t=ev(t)}return n}let o=eL(e,t);for(;o&&["table","td","th"].includes(ei(o))&&eA(o);)o=eL(o,t);return o&&eg(o)&&eA(o)&&!ep(o)?n:o||function(e){let t=ev(e);for(;eu(t)&&!eg(t);){if(ep(t))return t;if(ef(t))break;t=ev(t)}return null}(e)||n}let eI=async function(e){let t=this.getOffsetParent||eM,n=this.getDimensions,o=await n(e.floating);return{reference:function(e,t,n){let o=eu(t),r=es(t),i="fixed"===n,l=e_(e,!0,i,t),s={scrollLeft:0,scrollTop:0},a=H(0);if(o||!o&&!i){if(("body"!==ei(t)||eh(r))&&(s=eb(t)),o){let e=e_(t,!0,i,t);a.x=e.x+t.clientLeft,a.y=e.y+t.clientTop}else r&&(a.x=eC(r))}let c=!r||o||i?H(0):eO(r,s);return{x:l.left+s.scrollLeft-a.x-c.x,y:l.top+s.scrollTop-a.y-c.y,width:l.width,height:l.height}}(e.reference,await t(e.floating),e.strategy),floating:{x:0,y:0,width:o.width,height:o.height}}},eR={convertOffsetParentRelativeRectToViewportRelativeRect:function(e){let{elements:t,rect:n,offsetParent:o,strategy:r}=e,i="fixed"===r,l=es(o),s=!!t&&ef(t.floating);if(o===l||s&&i)return n;let a={scrollLeft:0,scrollTop:0},c=H(1),u=H(0),d=eu(o);if((d||!d&&!i)&&(("body"!==ei(o)||eh(l))&&(a=eb(o)),eu(o))){let e=e_(o);c=eT(o),u.x=e.x+o.clientLeft,u.y=e.y+o.clientTop}let h=!l||d||i?H(0):eO(l,a,!0);return{width:n.width*c.x,height:n.height*c.y,x:n.x*c.x-a.scrollLeft*c.x+u.x+h.x,y:n.y*c.y-a.scrollTop*c.y+u.y+h.y}},getDocumentElement:es,getClippingRect:function(e){let{element:t,boundary:n,rootBoundary:o,strategy:r}=e,i=[..."clippingAncestors"===n?ef(t)?[]:function(e,t){let n=t.get(e);if(n)return n;let o=ex(e,[],!1).filter(e=>ec(e)&&"body"!==ei(e)),r=null,i="fixed"===ey(e).position,l=i?ev(e):e;for(;ec(l)&&!eg(l);){let t=ey(l),n=ep(l);n||"fixed"!==t.position||(r=null),(i?!n&&!r:!n&&"static"===t.position&&!!r&&["absolute","fixed"].includes(r.position)||eh(l)&&!n&&function e(t,n){let o=ev(t);return!(o===n||!ec(o)||eg(o))&&("fixed"===ey(o).position||e(o,n))}(e,l))?o=o.filter(e=>e!==l):r=t,l=ev(l)}return t.set(e,o),o}(t,this._c):[].concat(n),o],l=i[0],s=i.reduce((e,n)=>{let o=eP(t,n,r);return e.top=R(o.top,e.top),e.right=I(o.right,e.right),e.bottom=I(o.bottom,e.bottom),e.left=R(o.left,e.left),e},eP(t,l,r));return{width:s.right-s.left,height:s.bottom-s.top,x:s.left,y:s.top}},getOffsetParent:eM,getElementRects:eI,getClientRects:function(e){return Array.from(e.getClientRects())},getDimensions:function(e){let{width:t,height:n}=eE(e);return{width:t,height:n}},getScale:eT,isElement:ec,isRTL:function(e){return"rtl"===ey(e).direction}},eq=e=>({name:"arrow",options:e,async fn(t){let{x:n,y:o,placement:r,rects:i,platform:l,elements:s,middlewareData:a}=t,{element:c,padding:u=0}=F(e,t)||{};if(null==c)return{};let d=J(u),h={x:n,y:o},f=W(U(r)),p=z(f),g=await l.getDimensions(c),y="y"===f,b=y?"clientHeight":"clientWidth",v=i.reference[p]+i.reference[f]-h[f]-i.floating[p],x=h[f]-i.reference[f],w=await (null==l.getOffsetParent?void 0:l.getOffsetParent(c)),E=w?w[b]:0;E&&await (null==l.isElement?void 0:l.isElement(w))||(E=s.floating[b]||i.floating[p]);let S=E/2-g[p]/2-1,T=I(d[y?"top":"left"],S),$=I(d[y?"bottom":"right"],S),k=E-g[p]-$,_=E/2-g[p]/2+(v/2-x/2),C=R(T,I(_,k)),O=!a.arrow&&null!=V(r)&&_!==C&&i.reference[p]/2-(_<T?T:$)-g[p]/2<0,P=O?_<T?_-T:_-k:0;return{[f]:h[f]+P,data:m({[f]:C,centerOffset:_-C-P},O&&{alignmentOffset:P}),reset:O}}}),eB=(e,t,n)=>{let o=new Map,r=m({platform:eR},n),i=m({},r.platform,{_c:o});return en(e,t,m({},r,{platform:i}))};function eH(){}function ej(e){return e()}function eD(){return Object.create(null)}function eF(e){e.forEach(ej)}function eN(e){return"function"==typeof e}function eV(e,t){return e!=e?t==t:e!==t||e&&"object"==typeof e||"function"==typeof e}function eW(e,t){e.appendChild(t)}function ez(e,t,n){e.insertBefore(t,n||null)}function eU(e){e.parentNode&&e.parentNode.removeChild(e)}function eY(e){return document.createElement(e)}function eG(e){return document.createElementNS("http://www.w3.org/2000/svg",e)}function eK(e){return document.createTextNode(e)}function eJ(){return eK(" ")}function eX(e,t,n,o){return e.addEventListener(t,n,o),()=>e.removeEventListener(t,n,o)}function eZ(e,t,n){null==n?e.removeAttribute(t):e.getAttribute(t)!==n&&e.setAttribute(t,n)}let eQ=["width","height"];function e0(e,t){let n=Object.getOwnPropertyDescriptors(e.__proto__);for(let o in t)null==t[o]?e.removeAttribute(o):"style"===o?e.style.cssText=t[o]:"__value"===o?e.value=e[o]=t[o]:n[o]&&n[o].set&&-1===eQ.indexOf(o)?e[o]=t[o]:eZ(e,o,t[o])}function e1(e,t,n){e.classList.toggle(t,!!n)}function e2(){if(!o)throw Error("Function called outside component initialization");return o}function e4(e){e2().$$.after_update.push(e)}let e6=[],e3=[],e5=[],e9=[],e8=Promise.resolve(),e7=!1;function te(e){e5.push(e)}let tt=new Set,tn=0;function to(){if(0!==tn)return;let e=o;do{try{for(;tn<e6.length;){let e=e6[tn];tn++,o=e,function(e){if(null!==e.fragment){e.update(),eF(e.before_update);let t=e.dirty;e.dirty=[-1],e.fragment&&e.fragment.p(e.ctx,t),e.after_update.forEach(te)}}(e.$$)}}catch(e){throw e6.length=0,tn=0,e}for(o=null,e6.length=0,tn=0;e3.length;)e3.pop()();for(let e=0;e<e5.length;e+=1){let t=e5[e];tt.has(t)||(tt.add(t),t())}e5.length=0}while(e6.length);for(;e9.length;)e9.pop()();e7=!1,tt.clear(),o=e}let tr=new Set;function ti(){r={r:0,c:[],p:r}}function tl(){r.r||eF(r.c),r=r.p}function ts(e,t){e&&e.i&&(tr.delete(e),e.i(t))}function ta(e,t,n,o){e&&e.o?tr.has(e)||(tr.add(e),r.c.push(()=>{tr.delete(e),o&&(n&&e.d(1),o())}),e.o(t)):o&&o()}function tc(e){return(null==e?void 0:e.length)!==void 0?e:Array.from(e)}function tu(e){e&&e.c()}function td(e,t,n){let{fragment:o,after_update:r}=e.$$;o&&o.m(t,n),te(()=>{let t=e.$$.on_mount.map(ej).filter(eN);e.$$.on_destroy?e.$$.on_destroy.push(...t):eF(t),e.$$.on_mount=[]}),r.forEach(te)}function th(e,t){let n=e.$$;null!==n.fragment&&(!function(e){let t=[],n=[];e5.forEach(o=>-1===e.indexOf(o)?t.push(o):n.push(o)),n.forEach(e=>e()),e5=t}(n.after_update),eF(n.on_destroy),n.fragment&&n.fragment.d(t),n.on_destroy=n.fragment=null,n.ctx=[])}function tf(e,t,n,r,i,l,s=null,a=[-1]){let c=o;o=e;let u=e.$$={fragment:null,ctx:[],props:l,update:eH,not_equal:i,bound:eD(),on_mount:[],on_destroy:[],on_disconnect:[],before_update:[],after_update:[],context:new Map(t.context||(c?c.$$.context:[])),callbacks:eD(),dirty:a,skip_bound:!1,root:t.target||c.$$.root};s&&s(u.root);let d=!1;if(u.ctx=n?n(e,t.props||{},(t,n,...o)=>{let r=o.length?o[0]:n;return u.ctx&&i(u.ctx[t],u.ctx[t]=r)&&(!u.skip_bound&&u.bound[t]&&u.bound[t](r),d)&&(-1===e.$$.dirty[0]&&(e6.push(e),e7||(e7=!0,e8.then(to)),e.$$.dirty.fill(0)),e.$$.dirty[t/31|0]|=1<<t%31),n}):[],u.update(),d=!0,eF(u.before_update),u.fragment=!!r&&r(u.ctx),t.target){if(t.hydrate){let e=Array.from(t.target.childNodes);u.fragment&&u.fragment.l(e),e.forEach(eU)}else u.fragment&&u.fragment.c();t.intro&&ts(e.$$.fragment),td(e,t.target,t.anchor),to()}o=c}class tp{constructor(){this.$$=void 0,this.$$set=void 0}$destroy(){th(this,1),this.$destroy=eH}$on(e,t){if(!eN(t))return eH;let n=this.$$.callbacks[e]||(this.$$.callbacks[e]=[]);return n.push(t),()=>{let e=n.indexOf(t);-1!==e&&n.splice(e,1)}}$set(e){this.$$set&&0!==Object.keys(e).length&&(this.$$.skip_bound=!0,this.$$set(e),this.$$.skip_bound=!1)}}function tm(e){let t,n,o,r,i;return{c(){eZ(t=eY("button"),"aria-label",n=e[3]?e[3]:null),eZ(t,"class",o=`${e[1]||""} shepherd-button ${e[4]?"shepherd-button-secondary":""}`),t.disabled=e[2],eZ(t,"tabindex","0"),eZ(t,"type","button")},m(n,o){ez(n,t,o),t.innerHTML=e[5],r||(i=eX(t,"click",function(){eN(e[0])&&e[0].apply(this,arguments)}),r=!0)},p(r,[i]){e=r,32&i&&(t.innerHTML=e[5]),8&i&&n!==(n=e[3]?e[3]:null)&&eZ(t,"aria-label",n),18&i&&o!==(o=`${e[1]||""} shepherd-button ${e[4]?"shepherd-button-secondary":""}`)&&eZ(t,"class",o),4&i&&(t.disabled=e[2])},i:eH,o:eH,d(e){e&&eU(t),r=!1,i()}}}function tg(e,t,n){let o,r,i,l,s,a,{config:c,step:u}=t;function h(e){return d(e)?e=e.call(u):e}return e.$$set=e=>{"config"in e&&n(6,c=e.config),"step"in e&&n(7,u=e.step)},e.$$.update=()=>{192&e.$$.dirty&&(n(0,o=c.action?c.action.bind(u.tour):null),n(1,r=c.classes),n(2,i=!!c.disabled&&h(c.disabled)),n(3,l=c.label?h(c.label):null),n(4,s=c.secondary),n(5,a=c.text?h(c.text):null))},[o,r,i,l,s,a,c,u]}"undefined"!=typeof window&&(window.__svelte||(window.__svelte={v:new Set})).v.add("4");class ty extends tp{constructor(e){super(),tf(this,e,tg,tm,eV,{config:6,step:7})}}function tb(e,t,n){let o=e.slice();return o[2]=t[n],o}function tv(e){let t,n;let o=tc(e[1]),r=[];for(let t=0;t<o.length;t+=1)r[t]=tx(tb(e,o,t));let i=e=>ta(r[e],1,1,()=>{r[e]=null});return{c(){for(let e=0;e<r.length;e+=1)r[e].c();t=eK("")},m(e,o){for(let t=0;t<r.length;t+=1)r[t]&&r[t].m(e,o);ez(e,t,o),n=!0},p(e,n){if(3&n){let l;for(l=0,o=tc(e[1]);l<o.length;l+=1){let i=tb(e,o,l);r[l]?(r[l].p(i,n),ts(r[l],1)):(r[l]=tx(i),r[l].c(),ts(r[l],1),r[l].m(t.parentNode,t))}for(ti(),l=o.length;l<r.length;l+=1)i(l);tl()}},i(e){if(!n){for(let e=0;e<o.length;e+=1)ts(r[e]);n=!0}},o(e){r=r.filter(Boolean);for(let e=0;e<r.length;e+=1)ta(r[e]);n=!1},d(e){e&&eU(t),function(e,t){for(let n=0;n<e.length;n+=1)e[n]&&e[n].d(t)}(r,e)}}}function tx(e){let t,n;return t=new ty({props:{config:e[2],step:e[0]}}),{c(){tu(t.$$.fragment)},m(e,o){td(t,e,o),n=!0},p(e,n){let o={};2&n&&(o.config=e[2]),1&n&&(o.step=e[0]),t.$set(o)},i(e){n||(ts(t.$$.fragment,e),n=!0)},o(e){ta(t.$$.fragment,e),n=!1},d(e){th(t,e)}}}function tw(e){let t,n;let o=e[1]&&tv(e);return{c(){t=eY("footer"),o&&o.c(),eZ(t,"class","shepherd-footer")},m(e,r){ez(e,t,r),o&&o.m(t,null),n=!0},p(e,[n]){e[1]?o?(o.p(e,n),2&n&&ts(o,1)):((o=tv(e)).c(),ts(o,1),o.m(t,null)):o&&(ti(),ta(o,1,1,()=>{o=null}),tl())},i(e){n||(ts(o),n=!0)},o(e){ta(o),n=!1},d(e){e&&eU(t),o&&o.d()}}}function tE(e,t,n){let o;let{step:r}=t;return e.$$set=e=>{"step"in e&&n(0,r=e.step)},e.$$.update=()=>{1&e.$$.dirty&&n(1,o=r.options.buttons)},[r,o]}class tS extends tp{constructor(e){super(),tf(this,e,tE,tw,eV,{step:0})}}function tT(e){let t,n,o,r,i;return{c(){t=eY("button"),(n=eY("span")).textContent="\xd7",eZ(n,"aria-hidden","true"),eZ(t,"aria-label",o=e[0].label?e[0].label:"Close Tour"),eZ(t,"class","shepherd-cancel-icon"),eZ(t,"type","button")},m(o,l){ez(o,t,l),eW(t,n),r||(i=eX(t,"click",e[1]),r=!0)},p(e,[n]){1&n&&o!==(o=e[0].label?e[0].label:"Close Tour")&&eZ(t,"aria-label",o)},i:eH,o:eH,d(e){e&&eU(t),r=!1,i()}}}function t$(e,t,n){let{cancelIcon:o,step:r}=t;return e.$$set=e=>{"cancelIcon"in e&&n(0,o=e.cancelIcon),"step"in e&&n(2,r=e.step)},[o,e=>{e.preventDefault(),r.cancel()},r]}class tk extends tp{constructor(e){super(),tf(this,e,t$,tT,eV,{cancelIcon:0,step:2})}}function t_(e){let t;return{c(){eZ(t=eY("h3"),"id",e[1]),eZ(t,"class","shepherd-title")},m(n,o){ez(n,t,o),e[3](t)},p(e,[n]){2&n&&eZ(t,"id",e[1])},i:eH,o:eH,d(n){n&&eU(t),e[3](null)}}}function tC(e,t,n){let{labelId:o,element:r,title:i}=t;return e4(()=>{d(i)&&n(2,i=i()),n(0,r.innerHTML=i,r)}),e.$$set=e=>{"labelId"in e&&n(1,o=e.labelId),"element"in e&&n(0,r=e.element),"title"in e&&n(2,i=e.title)},[r,o,i,function(e){e3[e?"unshift":"push"](()=>{n(0,r=e)})}]}class tO extends tp{constructor(e){super(),tf(this,e,tC,t_,eV,{labelId:1,element:0,title:2})}}function tP(e){let t,n;return t=new tO({props:{labelId:e[0],title:e[2]}}),{c(){tu(t.$$.fragment)},m(e,o){td(t,e,o),n=!0},p(e,n){let o={};1&n&&(o.labelId=e[0]),4&n&&(o.title=e[2]),t.$set(o)},i(e){n||(ts(t.$$.fragment,e),n=!0)},o(e){ta(t.$$.fragment,e),n=!1},d(e){th(t,e)}}}function tA(e){let t,n;return t=new tk({props:{cancelIcon:e[3],step:e[1]}}),{c(){tu(t.$$.fragment)},m(e,o){td(t,e,o),n=!0},p(e,n){let o={};8&n&&(o.cancelIcon=e[3]),2&n&&(o.step=e[1]),t.$set(o)},i(e){n||(ts(t.$$.fragment,e),n=!0)},o(e){ta(t.$$.fragment,e),n=!1},d(e){th(t,e)}}}function tL(e){let t,n,o;let r=e[2]&&tP(e),i=e[3]&&e[3].enabled&&tA(e);return{c(){t=eY("header"),r&&r.c(),n=eJ(),i&&i.c(),eZ(t,"class","shepherd-header")},m(e,l){ez(e,t,l),r&&r.m(t,null),eW(t,n),i&&i.m(t,null),o=!0},p(e,[o]){e[2]?r?(r.p(e,o),4&o&&ts(r,1)):((r=tP(e)).c(),ts(r,1),r.m(t,n)):r&&(ti(),ta(r,1,1,()=>{r=null}),tl()),e[3]&&e[3].enabled?i?(i.p(e,o),8&o&&ts(i,1)):((i=tA(e)).c(),ts(i,1),i.m(t,null)):i&&(ti(),ta(i,1,1,()=>{i=null}),tl())},i(e){o||(ts(r),ts(i),o=!0)},o(e){ta(r),ta(i),o=!1},d(e){e&&eU(t),r&&r.d(),i&&i.d()}}}function tM(e,t,n){let o,r,{labelId:i,step:l}=t;return e.$$set=e=>{"labelId"in e&&n(0,i=e.labelId),"step"in e&&n(1,l=e.step)},e.$$.update=()=>{2&e.$$.dirty&&(n(2,o=l.options.title),n(3,r=l.options.cancelIcon))},[i,l,o,r]}class tI extends tp{constructor(e){super(),tf(this,e,tM,tL,eV,{labelId:0,step:1})}}function tR(e){let t;return{c(){eZ(t=eY("div"),"class","shepherd-text"),eZ(t,"id",e[1])},m(n,o){ez(n,t,o),e[3](t)},p(e,[n]){2&n&&eZ(t,"id",e[1])},i:eH,o:eH,d(n){n&&eU(t),e[3](null)}}}function tq(e,t,n){let{descriptionId:o,element:r,step:i}=t;return e4(()=>{let{text:e}=i.options;d(e)&&(e=e.call(i)),u(e)?r.appendChild(e):n(0,r.innerHTML=e,r)}),e.$$set=e=>{"descriptionId"in e&&n(1,o=e.descriptionId),"element"in e&&n(0,r=e.element),"step"in e&&n(2,i=e.step)},[r,o,i,function(e){e3[e?"unshift":"push"](()=>{n(0,r=e)})}]}class tB extends tp{constructor(e){super(),tf(this,e,tq,tR,eV,{descriptionId:1,element:0,step:2})}}function tH(e){let t,n;return t=new tI({props:{labelId:e[1],step:e[2]}}),{c(){tu(t.$$.fragment)},m(e,o){td(t,e,o),n=!0},p(e,n){let o={};2&n&&(o.labelId=e[1]),4&n&&(o.step=e[2]),t.$set(o)},i(e){n||(ts(t.$$.fragment,e),n=!0)},o(e){ta(t.$$.fragment,e),n=!1},d(e){th(t,e)}}}function tj(e){let t,n;return t=new tB({props:{descriptionId:e[0],step:e[2]}}),{c(){tu(t.$$.fragment)},m(e,o){td(t,e,o),n=!0},p(e,n){let o={};1&n&&(o.descriptionId=e[0]),4&n&&(o.step=e[2]),t.$set(o)},i(e){n||(ts(t.$$.fragment,e),n=!0)},o(e){ta(t.$$.fragment,e),n=!1},d(e){th(t,e)}}}function tD(e){let t,n;return t=new tS({props:{step:e[2]}}),{c(){tu(t.$$.fragment)},m(e,o){td(t,e,o),n=!0},p(e,n){let o={};4&n&&(o.step=e[2]),t.$set(o)},i(e){n||(ts(t.$$.fragment,e),n=!0)},o(e){ta(t.$$.fragment,e),n=!1},d(e){th(t,e)}}}function tF(e){let t,n,o,r;let i=!f(e[2].options.title)||e[2].options.cancelIcon&&e[2].options.cancelIcon.enabled,l=!f(e[2].options.text),s=Array.isArray(e[2].options.buttons)&&e[2].options.buttons.length,a=i&&tH(e),c=l&&tj(e),u=s&&tD(e);return{c(){t=eY("div"),a&&a.c(),n=eJ(),c&&c.c(),o=eJ(),u&&u.c(),eZ(t,"class","shepherd-content")},m(e,i){ez(e,t,i),a&&a.m(t,null),eW(t,n),c&&c.m(t,null),eW(t,o),u&&u.m(t,null),r=!0},p(e,[r]){4&r&&(i=!f(e[2].options.title)||e[2].options.cancelIcon&&e[2].options.cancelIcon.enabled),i?a?(a.p(e,r),4&r&&ts(a,1)):((a=tH(e)).c(),ts(a,1),a.m(t,n)):a&&(ti(),ta(a,1,1,()=>{a=null}),tl()),4&r&&(l=!f(e[2].options.text)),l?c?(c.p(e,r),4&r&&ts(c,1)):((c=tj(e)).c(),ts(c,1),c.m(t,o)):c&&(ti(),ta(c,1,1,()=>{c=null}),tl()),4&r&&(s=Array.isArray(e[2].options.buttons)&&e[2].options.buttons.length),s?u?(u.p(e,r),4&r&&ts(u,1)):((u=tD(e)).c(),ts(u,1),u.m(t,null)):u&&(ti(),ta(u,1,1,()=>{u=null}),tl())},i(e){r||(ts(a),ts(c),ts(u),r=!0)},o(e){ta(a),ta(c),ta(u),r=!1},d(e){e&&eU(t),a&&a.d(),c&&c.d(),u&&u.d()}}}function tN(e,t,n){let{descriptionId:o,labelId:r,step:i}=t;return e.$$set=e=>{"descriptionId"in e&&n(0,o=e.descriptionId),"labelId"in e&&n(1,r=e.labelId),"step"in e&&n(2,i=e.step)},[o,r,i]}class tV extends tp{constructor(e){super(),tf(this,e,tN,tF,eV,{descriptionId:0,labelId:1,step:2})}}function tW(e){let t;return{c(){eZ(t=eY("div"),"class","shepherd-arrow"),eZ(t,"data-popper-arrow","")},m(e,n){ez(e,t,n)},d(e){e&&eU(t)}}}function tz(e){let t,n,o,r,i,l,s,a;let c=e[4].options.arrow&&e[4].options.attachTo&&e[4].options.attachTo.element&&e[4].options.attachTo.on&&tW();o=new tV({props:{descriptionId:e[2],labelId:e[3],step:e[4]}});let u=[{"aria-describedby":r=f(e[4].options.text)?null:e[2]},{"aria-labelledby":i=e[4].options.title?e[3]:null},e[1],{open:"true"}],d={};for(let e=0;e<u.length;e+=1)d=function(e,t){for(let n in t)e[n]=t[n];return e}(d,u[e]);return{c(){t=eY("dialog"),c&&c.c(),n=eJ(),tu(o.$$.fragment),e0(t,d),e1(t,"shepherd-has-cancel-icon",e[5]),e1(t,"shepherd-has-title",e[6]),e1(t,"shepherd-element",!0)},m(r,i){ez(r,t,i),c&&c.m(t,null),eW(t,n),td(o,t,null),e[13](t),l=!0,s||(a=eX(t,"keydown",e[7]),s=!0)},p(e,[s]){e[4].options.arrow&&e[4].options.attachTo&&e[4].options.attachTo.element&&e[4].options.attachTo.on?c||((c=tW()).c(),c.m(t,n)):c&&(c.d(1),c=null);let a={};4&s&&(a.descriptionId=e[2]),8&s&&(a.labelId=e[3]),16&s&&(a.step=e[4]),o.$set(a),e0(t,d=function(e,t){let n={},o={},r={$$scope:1},i=e.length;for(;i--;){let l=e[i],s=t[i];if(s){for(let e in l)e in s||(o[e]=1);for(let e in s)r[e]||(n[e]=s[e],r[e]=1);e[i]=s}else for(let e in l)r[e]=1}for(let e in o)e in n||(n[e]=void 0);return n}(u,[(!l||20&s&&r!==(r=f(e[4].options.text)?null:e[2]))&&{"aria-describedby":r},(!l||24&s&&i!==(i=e[4].options.title?e[3]:null))&&{"aria-labelledby":i},2&s&&e[1],{open:"true"}])),e1(t,"shepherd-has-cancel-icon",e[5]),e1(t,"shepherd-has-title",e[6]),e1(t,"shepherd-element",!0)},i(e){l||(ts(o.$$.fragment,e),l=!0)},o(e){ta(o.$$.fragment,e),l=!1},d(n){n&&eU(t),c&&c.d(),th(o),e[13](null),s=!1,a()}}}function tU(e){return e.split(" ").filter(e=>!!e.length)}function tY(e,t,n){var o;let r,i,l,{classPrefix:s,element:a,descriptionId:c,firstFocusableElement:u,focusableElements:d,labelId:f,lastFocusableElement:p,step:m,dataStepId:g}=t;return o=()=>{n(1,g={[`data-${s}shepherd-step-id`]:m.id}),n(9,d=a.querySelectorAll('a[href], area[href], input:not([disabled]), select:not([disabled]), textarea:not([disabled]), button:not([disabled]), [tabindex="0"]')),n(8,u=d[0]),n(10,p=d[d.length-1])},e2().$$.on_mount.push(o),e4(()=>{l!==m.options.classes&&(function(e){if(h(e)){let t=tU(e);t.length&&a.classList.remove(...t)}}(l),function(e){if(h(e)){let t=tU(e);t.length&&a.classList.add(...t)}}(l=m.options.classes))}),e.$$set=e=>{"classPrefix"in e&&n(11,s=e.classPrefix),"element"in e&&n(0,a=e.element),"descriptionId"in e&&n(2,c=e.descriptionId),"firstFocusableElement"in e&&n(8,u=e.firstFocusableElement),"focusableElements"in e&&n(9,d=e.focusableElements),"labelId"in e&&n(3,f=e.labelId),"lastFocusableElement"in e&&n(10,p=e.lastFocusableElement),"step"in e&&n(4,m=e.step),"dataStepId"in e&&n(1,g=e.dataStepId)},e.$$.update=()=>{16&e.$$.dirty&&(n(5,r=m.options&&m.options.cancelIcon&&m.options.cancelIcon.enabled),n(6,i=m.options&&m.options.title))},[a,g,c,f,m,r,i,e=>{let{tour:t}=m;switch(e.keyCode){case 9:if(0===d.length){e.preventDefault();break}e.shiftKey?(document.activeElement===u||document.activeElement.classList.contains("shepherd-element"))&&(e.preventDefault(),p.focus()):document.activeElement===p&&(e.preventDefault(),u.focus());break;case 27:t.options.exitOnEsc&&(e.preventDefault(),e.stopPropagation(),m.cancel());break;case 37:t.options.keyboardNavigation&&(e.preventDefault(),e.stopPropagation(),t.back());break;case 39:t.options.keyboardNavigation&&(e.preventDefault(),e.stopPropagation(),t.next())}},u,d,p,s,()=>a,function(e){e3[e?"unshift":"push"](()=>{n(0,a=e)})}]}class tG extends tp{constructor(e){super(),tf(this,e,tY,tz,eV,{classPrefix:11,element:0,descriptionId:2,firstFocusableElement:8,focusableElements:9,labelId:3,lastFocusableElement:10,step:4,dataStepId:1,getElement:12})}get getElement(){return this.$$.ctx[12]}}class tK extends p{constructor(e,t={}){return super(),this._resolvedAttachTo=void 0,this._resolvedExtraHighlightElements=void 0,this.classPrefix=void 0,this.el=void 0,this.target=void 0,this.tour=void 0,this.tour=e,this.classPrefix=this.tour.options?P(this.tour.options.classPrefix):"",this.styles=e.styles,this._resolvedAttachTo=null,_(this),this._setOptions(t),this}cancel(){this.tour.cancel(),this.trigger("cancel")}complete(){this.tour.complete(),this.trigger("complete")}destroy(){this.cleanup&&this.cleanup(),this.cleanup=null,u(this.el)&&(this.el.remove(),this.el=null),this._updateStepTargetOnHide(),this.trigger("destroy")}getTour(){return this.tour}hide(){var e;null==(e=this.tour.modal)||e.hide(),this.trigger("before-hide"),this.el&&(this.el.hidden=!0),this._updateStepTargetOnHide(),this.trigger("hide")}_resolveExtraHiglightElements(){return this._resolvedExtraHighlightElements=this.options.extraHighlights?this.options.extraHighlights.flatMap(e=>Array.from(document.querySelectorAll(e))):[],this._resolvedExtraHighlightElements}_resolveAttachToOptions(){return this._resolvedAttachTo=function(e){let t=e.options.attachTo||{},n=Object.assign({},t);if(d(n.element)&&(n.element=n.element.call(e)),h(n.element)){try{n.element=document.querySelector(n.element)}catch(e){}n.element||console.error(`The element for this Shepherd step was not found ${t.element}`)}return n}(this),this._resolvedAttachTo}_getResolvedAttachToOptions(){return null===this._resolvedAttachTo?this._resolveAttachToOptions():this._resolvedAttachTo}isOpen(){return!!(this.el&&!this.el.hidden)}show(){return d(this.options.beforeShowPromise)?Promise.resolve(this.options.beforeShowPromise()).then(()=>this._show()):Promise.resolve(this._show())}updateStepOptions(e){Object.assign(this.options,e),this.shepherdElementComponent&&this.shepherdElementComponent.$set({step:this})}getElement(){return this.el}getTarget(){return this.target}_createTooltipContent(){let e=`${this.id}-description`,t=`${this.id}-label`;return this.shepherdElementComponent=new tG({target:this.tour.options.stepsContainer||document.body,props:{classPrefix:this.classPrefix,descriptionId:e,labelId:t,step:this,styles:this.styles}}),this.shepherdElementComponent.getElement()}_scrollTo(e){let{element:t}=this._getResolvedAttachToOptions();d(this.options.scrollToHandler)?this.options.scrollToHandler(t):t instanceof Element&&"function"==typeof t.scrollIntoView&&t.scrollIntoView(e)}_getClassOptions(e){let t=this.tour&&this.tour.options&&this.tour.options.defaultStepOptions,n=e.classes?e.classes:"",o=t&&t.classes?t.classes:"";return Array.from(new Set([...n.split(" "),...o.split(" ")])).join(" ").trim()}_setOptions(e={}){let t=this.tour&&this.tour.options&&this.tour.options.defaultStepOptions;t=T({},t||{}),this.options=Object.assign({arrow:!0},t,e,{floatingUIOptions:T(t.floatingUIOptions||{},e.floatingUIOptions||{})});let{when:n}=this.options;this.options.classes=this._getClassOptions(e),this.destroy(),this.id=this.options.id||`step-${L()}`,n&&Object.keys(n).forEach(e=>{this.on(e,n[e],this)})}_setupElements(){f(this.el)||this.destroy(),this.el=this._createTooltipContent(),this.options.advanceOn&&function(e){let{event:t,selector:n}=e.options.advanceOn||{};if(!t)return console.error("advanceOn was defined, but no event name was passed.");{let o=t=>{if(e.isOpen()){let o=e.el&&t.currentTarget===e.el;(!f(n)&&t.currentTarget.matches(n)||o)&&e.tour.next()}},r=null;if(!f(n)&&!(r=document.querySelector(n)))return console.error(`No element was found for the selector supplied to advanceOn: ${n}`);r?(r.addEventListener(t,o),e.on("destroy",()=>r.removeEventListener(t,o))):(document.body.addEventListener(t,o,!0),e.on("destroy",()=>document.body.removeEventListener(t,o,!0)))}}(this),function(e){e.cleanup&&e.cleanup();let t=e._getResolvedAttachToOptions(),n=t.element,o=function(e,t){var n,o,r,i,l,s,a,c;let u={strategy:"absolute"};u.middleware=[];let d=!!t.options.arrow&&!!t.el&&t.el.querySelector(".shepherd-arrow"),h=A(e),f=null==(n=e.on)?void 0:n.includes("auto"),p=(null==e||null==(o=e.on)?void 0:o.includes("-start"))||(null==e||null==(r=e.on)?void 0:r.includes("-end"));if(!h){if(f?u.middleware.push({name:"autoPlacement",options:l={crossAxis:!0,alignment:p?null==e||null==(i=e.on)?void 0:i.split("-").pop():null},async fn(e){var t,n,o,r;let{rects:i,middlewareData:s,placement:a,platform:c,elements:u}=e,d=F(l,e),{crossAxis:h=!1,alignment:f,allowedPlacements:p=M,autoAlignment:m=!0}=d,y=g(d,Z),b=void 0!==f||p===M?((r=f||null)?[...p.filter(e=>V(e)===r),...p.filter(e=>V(e)!==r)]:p.filter(e=>N(e)===e)).filter(e=>!r||V(e)===r||!!m&&G(e)!==e):p,v=await eo(e,y),x=(null==(t=s.autoPlacement)?void 0:t.index)||0,w=b[x];if(null==w)return{};let E=Y(w,i,await (null==c.isRTL?void 0:c.isRTL(u.floating)));if(a!==w)return{reset:{placement:b[0]}};let S=[v[N(w)],v[E[0]],v[E[1]]],T=[...(null==(n=s.autoPlacement)?void 0:n.overflows)||[],{placement:w,overflows:S}],$=b[x+1];if($)return{data:{index:x+1,overflows:T},reset:{placement:$}};let k=T.map(e=>{let t=V(e.placement);return[e.placement,t&&h?e.overflows.slice(0,2).reduce((e,t)=>e+t,0):e.overflows[0],e.overflows]}).sort((e,t)=>e[1]-t[1]),_=(null==(o=k.filter(e=>e[2].slice(0,V(e[0])?2:3).every(e=>e<=0))[0])?void 0:o[0])||k[0][0];return _!==a?{data:{index:x+1,overflows:T},reset:{placement:_}}:{}}}):u.middleware.push((void 0===s&&(s={}),{name:"flip",options:s,async fn(e){var t,n,o,r,i;let{placement:l,middlewareData:a,rects:c,initialPlacement:u,platform:d,elements:h}=e,f=F(s,e),{mainAxis:p=!0,crossAxis:m=!0,fallbackPlacements:y,fallbackStrategy:b="bestFit",fallbackAxisSideDirection:v="none",flipAlignment:x=!0}=f,w=g(f,Q);if(null!=(t=a.arrow)&&t.alignmentOffset)return{};let E=N(l),S=U(u),T=N(u)===u,$=await (null==d.isRTL?void 0:d.isRTL(h.floating)),k=y||(T||!x?[K(u)]:function(e){let t=K(e);return[G(e),t,G(t)]}(u)),_="none"!==v;!y&&_&&k.push(...function(e,t,n,o){let r=V(e),i=function(e,t,n){let o=["left","right"],r=["right","left"];switch(e){case"top":case"bottom":if(n)return t?r:o;return t?o:r;case"left":case"right":return t?["top","bottom"]:["bottom","top"];default:return[]}}(N(e),"start"===n,o);return r&&(i=i.map(e=>e+"-"+r),t&&(i=i.concat(i.map(G)))),i}(u,x,v,$));let C=[u,...k],O=await eo(e,w),P=[],A=(null==(n=a.flip)?void 0:n.overflows)||[];if(p&&P.push(O[E]),m){let e=Y(l,c,$);P.push(O[e[0]],O[e[1]])}if(A=[...A,{placement:l,overflows:P}],!P.every(e=>e<=0)){let e=((null==(o=a.flip)?void 0:o.index)||0)+1,t=C[e];if(t)return{data:{index:e,overflows:A},reset:{placement:t}};let n=null==(r=A.filter(e=>e.overflows[0]<=0).sort((e,t)=>e.overflows[1]-t.overflows[1])[0])?void 0:r.placement;if(!n)switch(b){case"bestFit":{let e=null==(i=A.filter(e=>{if(_){let t=U(e.placement);return t===S||"y"===t}return!0}).map(e=>[e.placement,e.overflows.filter(e=>e>0).reduce((e,t)=>e+t,0)]).sort((e,t)=>e[1]-t[1])[0])?void 0:i[0];e&&(n=e);break}case"initialPlacement":n=u}if(l!==n)return{reset:{placement:n}}}return{}}})),u.middleware.push({name:"shift",options:c={limiter:(void 0===a&&(a={}),{options:a,fn(e){let{x:t,y:n,placement:o,rects:r,middlewareData:i}=e,{offset:l=0,mainAxis:s=!0,crossAxis:c=!0}=F(a,e),u={x:t,y:n},d=U(o),h=W(d),f=u[h],p=u[d],g=F(l,e),y="number"==typeof g?{mainAxis:g,crossAxis:0}:m({mainAxis:0,crossAxis:0},g);if(s){let e="y"===h?"height":"width",t=r.reference[h]-r.floating[e]+y.mainAxis,n=r.reference[h]+r.reference[e]-y.mainAxis;f<t?f=t:f>n&&(f=n)}if(c){var b,v;let e="y"===h?"width":"height",t=["top","left"].includes(N(o)),n=r.reference[d]-r.floating[e]+(t&&(null==(b=i.offset)?void 0:b[d])||0)+(t?0:y.crossAxis),l=r.reference[d]+r.reference[e]+(t?0:(null==(v=i.offset)?void 0:v[d])||0)-(t?y.crossAxis:0);p<n?p=n:p>l&&(p=l)}return{[h]:f,[d]:p}}}),crossAxis:!0},async fn(e){let{x:t,y:n,placement:o}=e,r=F(c,e),{mainAxis:i=!0,crossAxis:l=!1,limiter:s={fn:e=>{let{x:t,y:n}=e;return{x:t,y:n}}}}=r,a=g(r,ee),u={x:t,y:n},d=await eo(e,a),h=U(N(o)),f=W(h),p=u[f],y=u[h];if(i){let e="y"===f?"top":"left",t="y"===f?"bottom":"right",n=p+d[e],o=p-d[t];p=R(n,I(p,o))}if(l){let e="y"===h?"top":"left",t="y"===h?"bottom":"right",n=y+d[e],o=y-d[t];y=R(n,I(y,o))}let b=s.fn(m({},e,{[f]:p,[h]:y}));return m({},b,{data:{x:b.x-t,y:b.y-n}})}}),d){let e="object"==typeof t.options.arrow?t.options.arrow:{padding:4};u.middleware.push(eq({element:d,padding:p?e.padding:0}))}f||(u.placement=e.on)}return T(u,t.options.floatingUIOptions||{})}(t,e),r=A(t);r&&(n=document.body,e.shepherdElementComponent.getElement().classList.add("shepherd-centered")),e.cleanup=function(e,t,n,o){let r;void 0===o&&(o={});let{ancestorScroll:i=!0,ancestorResize:l=!0,elementResize:s="function"==typeof ResizeObserver,layoutShift:a="function"==typeof IntersectionObserver,animationFrame:c=!1}=o,u=eS(e),d=i||l?[...u?ex(u):[],...ex(t)]:[];d.forEach(e=>{i&&e.addEventListener("scroll",n,{passive:!0}),l&&e.addEventListener("resize",n)});let h=u&&a?function(e,t){let n,o=null,r=es(e);function i(){var e;clearTimeout(n),null==(e=o)||e.disconnect(),o=null}return function l(s,a){void 0===s&&(s=!1),void 0===a&&(a=1),i();let{left:c,top:u,width:d,height:h}=e.getBoundingClientRect();if(s||t(),!d||!h)return;let f=B(u),p=B(r.clientWidth-(c+d)),g={rootMargin:-f+"px "+-p+"px "+-B(r.clientHeight-(u+h))+"px "+-B(c)+"px",threshold:R(0,I(1,a))||1},y=!0;function b(e){let t=e[0].intersectionRatio;if(t!==a){if(!y)return l();t?l(!1,t):n=setTimeout(()=>{l(!1,1e-7)},1e3)}y=!1}try{o=new IntersectionObserver(b,m({},g,{root:r.ownerDocument}))}catch(e){o=new IntersectionObserver(b,g)}o.observe(e)}(!0),i}(u,n):null,f=-1,p=null;s&&(p=new ResizeObserver(e=>{let[o]=e;o&&o.target===u&&p&&(p.unobserve(t),cancelAnimationFrame(f),f=requestAnimationFrame(()=>{var e;null==(e=p)||e.observe(t)})),n()}),u&&!c&&p.observe(u),p.observe(t));let g=c?e_(e):null;return c&&function t(){let o=e_(e);g&&(o.x!==g.x||o.y!==g.y||o.width!==g.width||o.height!==g.height)&&n(),g=o,r=requestAnimationFrame(t)}(),n(),()=>{var e;d.forEach(e=>{i&&e.removeEventListener("scroll",n),l&&e.removeEventListener("resize",n)}),null==h||h(),null==(e=p)||e.disconnect(),p=null,c&&cancelAnimationFrame(r)}}(n,e.el,()=>{if(!e.el){null==e.cleanup||e.cleanup();return}(function(e,t,n,o){eB(e,t.el,n).then(({x:e,y:n,placement:r,middlewareData:i})=>(t.el&&(o?Object.assign(t.el.style,{position:"fixed",left:"50%",top:"50%",transform:"translate(-50%, -50%)"}):Object.assign(t.el.style,{position:"absolute",left:`${e}px`,top:`${n}px`}),t.el.dataset.popperPlacement=r,function(e,t){let n=e.querySelector(".shepherd-arrow");if(u(n)&&t.arrow){let{x:e,y:o}=t.arrow;Object.assign(n.style,{left:null!=e?`${e}px`:"",top:null!=o?`${o}px`:""})}}(t.el,i)),t)).then(e=>new Promise(t=>{setTimeout(()=>t(e),300)})).then(e=>{null!=e&&e.el&&e.el.focus({preventScroll:!0})})})(n,e,o,r)}),e.target=t.element}(this)}_show(){var e;this.trigger("before-show"),this._resolveAttachToOptions(),this._resolveExtraHiglightElements(),this._setupElements(),this.tour.modal||this.tour.setupModal(),null==(e=this.tour.modal)||e.setupForStep(this),this._styleTargetElementForStep(this),this.el&&(this.el.hidden=!1),this.options.scrollTo&&setTimeout(()=>{this._scrollTo(this.options.scrollTo)}),this.el&&(this.el.hidden=!1);let t=this.shepherdElementComponent.getElement(),n=this.target||document.body,o=this._resolvedExtraHighlightElements;n.classList.add(`${this.classPrefix}shepherd-enabled`),n.classList.add(`${this.classPrefix}shepherd-target`),t.classList.add("shepherd-enabled"),null==o||o.forEach(e=>{e.classList.add(`${this.classPrefix}shepherd-enabled`),e.classList.add(`${this.classPrefix}shepherd-target`)}),this.trigger("show")}_styleTargetElementForStep(e){let t=e.target,n=e._resolvedExtraHighlightElements;if(!t)return;let o=e.options.highlightClass;o&&(t.classList.add(o),null==n||n.forEach(e=>e.classList.add(o))),t.classList.remove("shepherd-target-click-disabled"),null==n||n.forEach(e=>e.classList.remove("shepherd-target-click-disabled")),!1===e.options.canClickTarget&&(t.classList.add("shepherd-target-click-disabled"),null==n||n.forEach(e=>e.classList.add("shepherd-target-click-disabled")))}_updateStepTargetOnHide(){let e=this.target||document.body,t=this._resolvedExtraHighlightElements,n=this.options.highlightClass;n&&(e.classList.remove(n),null==t||t.forEach(e=>e.classList.remove(n))),e.classList.remove("shepherd-target-click-disabled",`${this.classPrefix}shepherd-enabled`,`${this.classPrefix}shepherd-target`),null==t||t.forEach(e=>{e.classList.remove("shepherd-target-click-disabled",`${this.classPrefix}shepherd-enabled`,`${this.classPrefix}shepherd-target`)})}}function tJ(e){let t,n,o,r,i;return{c(){t=eG("svg"),eZ(n=eG("path"),"d",e[2]),eZ(t,"class",o=`${e[1]?"shepherd-modal-is-visible":""} shepherd-modal-overlay-container`)},m(o,l){ez(o,t,l),eW(t,n),e[11](t),r||(i=eX(t,"touchmove",e[3]),r=!0)},p(e,[r]){4&r&&eZ(n,"d",e[2]),2&r&&o!==(o=`${e[1]?"shepherd-modal-is-visible":""} shepherd-modal-overlay-container`)&&eZ(t,"class",o)},i:eH,o:eH,d(n){n&&eU(t),e[11](null),r=!1,i()}}}function tX(e,t,n){let o,r,{element:i,openingProperties:l}=t,s=!1;function a(){n(4,l=[{width:0,height:0,x:0,y:0,r:0}])}function c(){n(1,s=!1),f()}function u(e=0,t=0,o=0,r=0,i,s,c){if(s){let a=[s,...c||[]];for(let s of(n(4,l=[]),a)){if(!s||a.indexOf(s)!==a.lastIndexOf(s))continue;let{y:n,height:c}=function(e,t){let n=e.getBoundingClientRect(),o=n.y||n.top,r=n.bottom||o+n.height;if(t){let e=t.getBoundingClientRect(),n=e.y||e.top,i=e.bottom||n+e.height;o=Math.max(o,n),r=Math.min(r,i)}let i=Math.max(r-o,0);return{y:o,height:i}}(s,i),{x:u,width:d,left:h}=s.getBoundingClientRect();a.some(e=>{if(e===s)return!1;let t=e.getBoundingClientRect();return u>=t.left&&u+d<=t.right&&n>=t.top&&n+c<=t.bottom})||l.push({width:d+2*e,height:c+2*e,x:(u||h)+o-e,y:n+r-e,r:t})}}else a()}function d(){n(1,s=!0)}a();let h=e=>{e.preventDefault()};function f(){o&&(cancelAnimationFrame(o),o=void 0),window.removeEventListener("touchmove",h,{passive:!1})}return e.$$set=e=>{"element"in e&&n(0,i=e.element),"openingProperties"in e&&n(4,l=e.openingProperties)},e.$$.update=()=>{16&e.$$.dirty&&n(2,r=function(e){let t="",{innerWidth:n,innerHeight:o}=window;return e.forEach(e=>{let{width:n,height:o,x:r=0,y:i=0,r:l=0}=e,{topLeft:s=0,topRight:a=0,bottomRight:c=0,bottomLeft:u=0}="number"==typeof l?{topLeft:l,topRight:l,bottomRight:l,bottomLeft:l}:l;t+=`M${r+s},${i}\
      a${s},${s},0,0,0-${s},${s}\
      V${o+i-u}\
      a${u},${u},0,0,0,${u},${u}\
      H${n+r-c}\
      a${c},${c},0,0,0,${c}-${c}\
      V${i+a}\
      a${a},${a},0,0,0-${a}-${a}\
      Z`}),`M${n},${o}\
          H0\
          V0\
          H${n}\
          V${o}\
          Z\
          ${t}`.replace(/\s/g,"")}(l))},[i,s,r,e=>{e.stopPropagation()},l,()=>i,a,c,u,function(e){f(),e.tour.options.useModalOverlay?(function(e){let{modalOverlayOpeningPadding:t,modalOverlayOpeningRadius:n,modalOverlayOpeningXOffset:r=0,modalOverlayOpeningYOffset:i=0}=e.options,l=function(e){let t={top:0,left:0};if(!e)return t;let n=e.ownerDocument.defaultView;for(;n!==window.top;){var o,r,i;let e=null==(o=n)?void 0:o.frameElement;if(e){let n=e.getBoundingClientRect();t.top+=n.top+(null!=(r=n.scrollTop)?r:0),t.left+=n.left+(null!=(i=n.scrollLeft)?i:0)}n=n.parent}return t}(e.target),s=function e(t){if(!t)return null;let n=t instanceof HTMLElement&&window.getComputedStyle(t).overflowY;return"hidden"!==n&&"visible"!==n&&t.scrollHeight>=t.clientHeight?t:e(t.parentElement)}(e.target),a=()=>{o=void 0,u(t,n,r+l.left,i+l.top,s,e.target,e._resolvedExtraHighlightElements),o=requestAnimationFrame(a)};a(),window.addEventListener("touchmove",h,{passive:!1})}(e),d()):c()},d,function(e){e3[e?"unshift":"push"](()=>{n(0,i=e)})}]}class tZ extends tp{constructor(e){super(),tf(this,e,tX,tJ,eV,{element:0,openingProperties:4,getElement:5,closeModalOpening:6,hide:7,positionModal:8,setupForStep:9,show:10})}get getElement(){return this.$$.ctx[5]}get closeModalOpening(){return this.$$.ctx[6]}get hide(){return this.$$.ctx[7]}get positionModal(){return this.$$.ctx[8]}get setupForStep(){return this.$$.ctx[9]}get show(){return this.$$.ctx[10]}}class tQ extends p{constructor(){super(),this.activeTour=void 0,_(this)}}class t0 extends p{constructor(e={}){return super(),this.trackedEvents=["active","cancel","complete","show"],this.classPrefix=void 0,this.currentStep=void 0,this.focusedElBeforeOpen=void 0,this.id=void 0,this.modal=void 0,this.options=void 0,this.steps=void 0,_(this),this.options=Object.assign({},{exitOnEsc:!0,keyboardNavigation:!0},e),this.classPrefix=P(this.options.classPrefix),this.steps=[],this.addSteps(this.options.steps),["active","cancel","complete","inactive","show","start"].map(e=>{(e=>{this.on(e,t=>{(t=t||{}).tour=this,t1.trigger(e,t)})})(e)}),this._setTourID(e.id),this}addStep(e,t){let n=e;return n instanceof tK?n.tour=this:n=new tK(this,n),f(t)?this.steps.push(n):this.steps.splice(t,0,n),n}addSteps(e){return Array.isArray(e)&&e.forEach(e=>{this.addStep(e)}),this}back(){let e=this.steps.indexOf(this.currentStep);this.show(e-1,!1)}async cancel(){if(this.options.confirmCancel){let e=this.options.confirmCancelMessage||"Are you sure you want to stop the tour?";(d(this.options.confirmCancel)?await this.options.confirmCancel():window.confirm(e))&&this._done("cancel")}else this._done("cancel")}complete(){this._done("complete")}getById(e){return this.steps.find(t=>t.id===e)}getCurrentStep(){return this.currentStep}hide(){let e=this.getCurrentStep();if(e)return e.hide()}isActive(){return t1.activeTour===this}next(){let e=this.steps.indexOf(this.currentStep);e===this.steps.length-1?this.complete():this.show(e+1,!0)}removeStep(e){let t=this.getCurrentStep();this.steps.some((t,n)=>{if(t.id===e)return t.isOpen()&&t.hide(),t.destroy(),this.steps.splice(n,1),!0}),t&&t.id===e&&(this.currentStep=void 0,this.steps.length?this.show(0):this.cancel())}show(e=0,t=!0){let n=h(e)?this.getById(e):this.steps[e];n&&(this._updateStateBeforeShow(),d(n.options.showOn)&&!n.options.showOn()?this._skipStep(n,t):(this.currentStep=n,this.trigger("show",{step:n,previous:this.currentStep}),n.show()))}async start(){this.trigger("start"),this.focusedElBeforeOpen=document.activeElement,this.currentStep=null,this.setupModal(),this._setupActiveTour(),this.next()}_done(e){let t=this.steps.indexOf(this.currentStep);if(Array.isArray(this.steps)&&this.steps.forEach(e=>e.destroy()),!function(e){if(e){let{steps:t}=e;t.forEach(e=>{e.options&&!1===e.options.canClickTarget&&e.options.attachTo&&(u(e.target)&&e.target.classList.remove("shepherd-target-click-disabled"),e._resolvedExtraHighlightElements&&e._resolvedExtraHighlightElements.forEach(e=>{u(e)&&e.classList.remove("shepherd-target-click-disabled")}))})}}(this),this.trigger(e,{index:t}),t1.activeTour=null,this.trigger("inactive",{tour:this}),this.modal&&this.modal.hide(),("cancel"===e||"complete"===e)&&this.modal){let e=document.querySelector(".shepherd-modal-overlay-container");e&&(e.remove(),this.modal=null)}u(this.focusedElBeforeOpen)&&this.focusedElBeforeOpen.focus()}_setupActiveTour(){this.trigger("active",{tour:this}),t1.activeTour=this}setupModal(){this.modal=new tZ({target:this.options.modalContainer||document.body,props:{styles:this.styles}})}_skipStep(e,t){let n=this.steps.indexOf(e);n===this.steps.length-1?this.complete():this.show(t?n+1:n-1,t)}_updateStateBeforeShow(){this.currentStep&&this.currentStep.hide(),this.isActive()||this._setupActiveTour()}_setTourID(e){let t=this.options.tourName||"tour",n=e||L();this.id=`${t}--${n}`}}let t1=new tQ,t2="undefined"==typeof window;t1.Step=t2?C:tK,t1.Tour=t2?O:t0;var t4=n(6556);let t6=(0,a.createContext)(null);function t3(){let e=(0,a.useContext)(t6);if(!e)throw Error("useTour must be used within a TourProvider");return e}function t5(e){let{children:t}=e,n=(0,a.useRef)(null),o=(0,c.useRouter)(),{isFirstVisit:r,markTourCompleted:i}=(0,t4.n)(),[l,u]=(0,a.useState)(!1),[d,h]=(0,a.useState)(!1),[f,p]=(0,a.useState)(!1),m=(0,a.useRef)(!1),g=()=>{n.current&&(h(!0),p(!0),n.current.start(),setTimeout(()=>{p(!1)},1e3))};(0,a.useEffect)(()=>{if(!n.current){n.current=new t1.Tour({useModalOverlay:!1,defaultStepOptions:{cancelIcon:{enabled:!0},scrollTo:{behavior:"smooth",block:"center"},arrow:!1,highlightClass:"shepherd-highlight",when:{show(){var e,t,n;let o=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),r=null==o?void 0:o.getElement(),i=null==r?void 0:r.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(o))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),r){let e=r.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=r.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}let s=null==o?void 0:o.getTarget();s&&s instanceof HTMLElement&&(s.style.outline="3px solid #3b82f6",s.style.outlineOffset="2px",s.style.borderRadius="8px",s.style.position="relative",s.style.zIndex="9999",s.setAttribute("data-shepherd-highlighted","true"))},hide(){let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted")),document.documentElement.style.removeProperty("--shepherd-dialog-height");let t=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');t&&t instanceof HTMLElement&&(t.style.removeProperty("height"),t.style.removeProperty("max-height"))}}}});let e=document.createElement("style");e.id="shepherd-global-custom-style",e.textContent='\n          .shepherd-element {\n            /* Uniform 1px border using inner box-shadow so corners stay consistent */\n            border: none !important;\n            border-radius: 10px !important;\n            z-index: 30000 !important;\n            box-shadow: 0 0 0 1px #d1d5db inset, 0 10px 15px -3px rgba(0, 0, 0, 0.1), 0 4px 6px -2px rgba(0, 0, 0, 0.05) !important;\n            overflow: visible !important;\n            background-clip: padding-box !important;\n          }\n\n          .shepherd-title {\n              font-weight: bold;\n              color: #111827;\n              margin: 0;\n          }\n\n          .shepherd-element .shepherd-header {\n              padding: 1rem 1rem 0.5rem 1rem;\n          }\n\n          .shepherd-element .shepherd-text {\n              padding: 0.5rem 1rem 1rem 1rem;\n          }\n\n          /* Fix mobile menu gap when tour dialog is present */\n          @media (max-width: 767px) {\n            /* Very specific selector to override Tailwind\'s h-[calc(100vh-56px)] class */\n            div.fixed.top-14.left-0.w-64.bg-white.border-r.shadow-lg.z-50.transform,\n            .fixed.top-14.left-0.w-64.bg-white.shadow-lg.z-50,\n            div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"][class*="bg-white"][class*="shadow-lg"][class*="z-50"] {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n              max-height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Target the mobile menu by its exact class combination from the HTML */\n            .fixed.w-64.bg-white.border-r.border-gray-200.shadow-lg.z-50.transform,\n            .fixed[class*="w-64"][class*="bg-white"][class*="border-r"][class*="shadow-lg"][class*="z-50"][class*="transform"] {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Even more specific - target by multiple class combinations */\n            .fixed.top-14.left-0[class*="w-64"],\n            div.fixed[class*="top-14"][class*="left-0"][class*="w-64"] {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Super aggressive approach - use high specificity to override Tailwind */\n            body div.fixed.w-64:not(.shepherd-element),\n            html body div.fixed.w-64:not(.shepherd-element) {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Fallback selectors for other mobile menu patterns */\n            nav[data-state="open"],\n            .mobile-menu.open,\n            [data-mobile-menu="true"] {\n              height: calc(100vh - var(--shepherd-dialog-height, 200px)) !important;\n              max-height: calc(100vh - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Ensure mobile menu content flows properly */\n            .fixed.w-64 nav,\n            .fixed[class*="w-64"] nav {\n              height: 100% !important;\n              overflow-y: auto !important;\n            }\n          }\n\n        ',document.getElementById("shepherd-global-custom-style")||document.head.appendChild(e),n.current.on("complete",()=>{let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted"));let t=document.getElementById("shepherd-column-overlay");t&&t.remove();let n=document.getElementById("shepherd-column-anchor");n&&n.remove();let o=document.getElementById("shepherd-user-column-overlay");o&&o.remove();let r=document.getElementById("shepherd-user-column-anchor");r&&r.remove();let l=document.getElementById("shepherd-global-custom-style");l&&l.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let s=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');s&&s instanceof HTMLElement&&(s.style.removeProperty("height"),s.style.removeProperty("max-height")),h(!1),p(!1),i()}),n.current.on("cancel",()=>{let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted"));let t=document.getElementById("shepherd-column-overlay");t&&t.remove();let n=document.getElementById("shepherd-column-anchor");n&&n.remove();let o=document.getElementById("shepherd-user-column-overlay");o&&o.remove();let r=document.getElementById("shepherd-user-column-anchor");r&&r.remove();let l=document.getElementById("shepherd-global-custom-style");l&&l.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let s=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');s&&s instanceof HTMLElement&&(s.style.removeProperty("height"),s.style.removeProperty("max-height")),h(!1),p(!1),i()}),[{title:"\uD83D\uDC4B Welcome to SkyPilot!",text:"\n              <p>SkyPilot is a system for managing AI workloads on any cluster and cloud infrastructure.</p>\n            ",buttons:[{text:"Skip Tour",action(){this.cancel()},classes:"shepherd-button-secondary"},{text:"Start Tour",action(){this.next()}}]},{title:"Clusters",text:"\n              <p>Spin up <strong>Sky Clusters</strong> on any infrastructure you have access to.</p>\n              <p>Easily SSH into any node, connect an IDE, or queue development jobs.</p>\n            ",attachTo:{element:'a[href="/dashboard/clusters"]',on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{let t=()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()};"/dashboard/clusters"!==o.pathname?(m.current=!0,o.push("/clusters").then(()=>{m.current=!1,setTimeout(t,200)})):t()})},when:{show(){var e,t,n;let o=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),r=null==o?void 0:o.getElement(),i=null==r?void 0:r.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(o))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),r){let e=r.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=r.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}let s=document.querySelector('a[href="/dashboard/clusters"]');s&&s instanceof HTMLElement&&(s.style.outline="3px solid #3b82f6",s.style.outlineOffset="2px",s.style.borderRadius="8px",s.style.position="relative",s.style.zIndex="9999",s.setAttribute("data-shepherd-highlighted","true"))},hide(){let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted")),document.documentElement.style.removeProperty("--shepherd-dialog-height");let t=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');t&&t instanceof HTMLElement&&(t.style.removeProperty("height"),t.style.removeProperty("max-height"))}},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"SkyPilot is infra-agnostic",text:"\n              <p>Run compute on any hyperscaler, neocloud, or Kubernetes cluster — all within a unified system.</p>\n            ",attachTo:{element:function(){let e=document.getElementById("shepherd-column-anchor");if(e)return e;let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(t){let e=t.closest("table"),n=Array.from(t.parentElement.children).indexOf(t);if(e){let t=e.querySelectorAll("tbody tr"),o=null;for(let e=t.length-1;e>=0;e--){let r=t[e].children[n];if(r){o=r;break}}if(o)return o}return t}return document.querySelector("table")||"body"},on:"bottom",offset:{skidding:0,distance:15}},beforeShowPromise:function(){return new Promise(e=>{let t=()=>{let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(t&&t instanceof HTMLElement){let o=t.closest("table");if(o){let r=t.getBoundingClientRect(),i=window.innerWidth,l=o.closest('.overflow-x-auto, .overflow-auto, [style*="overflow"]')||o.parentElement;if((r.right>i||r.left<0)&&l&&l instanceof HTMLElement){let o=l.getBoundingClientRect(),i=t.offsetLeft-o.width/2+r.width/2;l.scrollTo({left:Math.max(0,i),behavior:"smooth"}),setTimeout(()=>{n(),e()},300);return}n()}}e()},n=()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),o=e.getBoundingClientRect(),r=t.querySelectorAll("tbody tr"),i=o;r.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=i.bottom+4,s=document.createElement("div");s.id="shepherd-column-anchor",s.style.position="fixed",s.style.left="".concat(o.left+o.width/2,"px"),s.style.top="".concat(l,"px"),s.style.width="1px",s.style.height="1px",s.style.zIndex="9999",s.style.pointerEvents="none",s.style.backgroundColor="transparent",s.style.transform="translate(-50%, -50%)",document.body.appendChild(s)}}};"/dashboard/clusters"!==window.location.pathname?(m.current=!0,o.push("/clusters").then(()=>{m.current=!1,setTimeout(t,200)})):t()})},when:{show(){var e,t,n;let o=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),r=null==o?void 0:o.getElement(),i=null==r?void 0:r.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(o))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),r){let e=r.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=r.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}setTimeout(()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),o=e.getBoundingClientRect(),r=t.querySelectorAll("tbody tr"),i=o;r.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=document.createElement("div");l.id="shepherd-column-overlay",l.style.position="fixed",l.style.left="".concat(o.left-4,"px"),l.style.top="".concat(o.top-4,"px"),l.style.width="".concat(o.width+8,"px"),l.style.height="".concat(i.bottom-o.top+8,"px"),l.style.outline="3px solid #3b82f6",l.style.outlineOffset="2px",l.style.borderRadius="8px",l.style.zIndex="9998",l.style.pointerEvents="none",l.style.backgroundColor="transparent",document.body.appendChild(l)}}},100)},hide(){let e=document.getElementById("shepherd-column-overlay");e&&e.remove();let t=document.getElementById("shepherd-column-anchor");t&&t.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let n=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');n&&n instanceof HTMLElement&&(n.style.removeProperty("height"),n.style.removeProperty("max-height"));let o=document.getElementById("shepherd-global-custom-style");o&&o.remove()}},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Multi-user support",text:"\n              <p>SkyPilot supports multiple users in an organization.</p>\n              <p>Each user can have their own clusters and jobs, with proper access controls.</p>\n            ",attachTo:{element:function(){let e=document.getElementById("shepherd-user-column-anchor");if(e)return e;let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(t){let e=t.closest("table"),n=Array.from(t.parentElement.children).indexOf(t);if(e){let t=e.querySelectorAll("tbody tr"),o=null;for(let e=t.length-1;e>=0;e--){let r=t[e].children[n];if(r){o=r;break}}if(o)return o}return t}return document.querySelector("table")||"body"},on:"bottom",offset:{skidding:0,distance:15}},beforeShowPromise:function(){return new Promise(e=>{let t=()=>{let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(t&&t instanceof HTMLElement){let o=t.closest("table");if(o){let r=t.getBoundingClientRect(),i=window.innerWidth,l=o.closest('.overflow-x-auto, .overflow-auto, [style*="overflow"]')||o.parentElement;if((r.right>i||r.left<0)&&l&&l instanceof HTMLElement){let o=l.getBoundingClientRect(),i=t.offsetLeft-o.width/2+r.width/2;l.scrollTo({left:Math.max(0,i),behavior:"smooth"}),setTimeout(()=>{n(),e()},300);return}n()}}e()},n=()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),o=e.getBoundingClientRect(),r=t.querySelectorAll("tbody tr"),i=o;r.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=i.bottom+4,s=document.createElement("div");s.id="shepherd-user-column-anchor",s.style.position="fixed",s.style.left="".concat(o.left+o.width/2,"px"),s.style.top="".concat(l,"px"),s.style.width="1px",s.style.height="1px",s.style.zIndex="9999",s.style.pointerEvents="none",s.style.backgroundColor="transparent",s.style.transform="translate(-50%, -50%)",document.body.appendChild(s)}}};"/dashboard/clusters"!==window.location.pathname?(m.current=!0,o.push("/clusters").then(()=>{m.current=!1,setTimeout(t,200)})):t()})},when:{show(){var e,t,n;let o=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),r=null==o?void 0:o.getElement(),i=null==r?void 0:r.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(o))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),r){let e=r.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=r.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}setTimeout(()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),o=e.getBoundingClientRect(),r=t.querySelectorAll("tbody tr"),i=o;r.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=document.createElement("div");l.id="shepherd-user-column-overlay",l.style.position="fixed",l.style.left="".concat(o.left-4,"px"),l.style.top="".concat(o.top-4,"px"),l.style.width="".concat(o.width+8,"px"),l.style.height="".concat(i.bottom-o.top+8,"px"),l.style.outline="3px solid #3b82f6",l.style.outlineOffset="2px",l.style.borderRadius="8px",l.style.zIndex="9998",l.style.pointerEvents="none",l.style.backgroundColor="transparent",document.body.appendChild(l)}}},100)},hide(){let e=document.getElementById("shepherd-user-column-overlay");e&&e.remove();let t=document.getElementById("shepherd-user-column-anchor");t&&t.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let n=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');n&&n instanceof HTMLElement&&(n.style.removeProperty("height"),n.style.removeProperty("max-height"));let o=document.getElementById("shepherd-global-custom-style");o&&o.remove()}},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Spin up compute in seconds",text:'\n            <p>Spin up clusters using the Python SDK or the CLI.</p>\n            <div class="space-y-2">\n              <div class="rounded-lg border text-card-foreground shadow-sm p-3 bg-gray-50">\n                <div class="flex items-center justify-between">\n                  <pre class="text-sm w-full whitespace-pre-wrap">\n                    <code class="block">sky launch</code>\n                  </pre>\n                  <button class="inline-flex items-center justify-center whitespace-nowrap text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 hover:bg-accent hover:text-accent-foreground h-8 w-8 rounded-full" onclick="copyDashboardCodeBlock(this)" title="Copy to clipboard">\n                    <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-copy h-4 w-4">\n                      <rect width="14" height="14" x="8" y="8" rx="2" ry="2"></rect>\n                      <path d="M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2"></path>\n                    </svg>\n                  </button>\n                </div>\n              </div>\n              <div class="rounded-lg border text-card-foreground shadow-sm p-3 bg-gray-50">\n                <div class="flex items-center justify-between">\n                  <pre class="text-sm w-full whitespace-pre-wrap">\n                    <code class="block">sky launch --gpus L4:8</code>\n                  </pre>\n                  <button class="inline-flex items-center justify-center whitespace-nowrap text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 hover:bg-accent hover:text-accent-foreground h-8 w-8 rounded-full" onclick="copyDashboardCodeBlock(this)" title="Copy to clipboard">\n                    <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-copy h-4 w-4">\n                      <rect width="14" height="14" x="8" y="8" rx="2" ry="2"></rect>\n                      <path d="M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2"></path>\n                    </svg>\n                  </button>\n                </div>\n              </div>\n              <div class="rounded-lg border text-card-foreground shadow-sm p-3 bg-gray-50">\n                <div class="flex items-center justify-between">\n                  <pre class="text-sm w-full whitespace-pre-wrap">\n                    <code class="block">sky launch --num-nodes 10 --cpus 32+</code>\n                  </pre>\n                  <button class="inline-flex items-center justify-center whitespace-nowrap text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 hover:bg-accent hover:text-accent-foreground h-8 w-8 rounded-full" onclick="copyDashboardCodeBlock(this)" title="Copy to clipboard">\n                    <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-copy h-4 w-4">\n                      <rect width="14" height="14" x="8" y="8" rx="2" ry="2"></rect>\n                      <path d="M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2"></path>\n                    </svg>\n                  </button>\n                </div>\n              </div>\n            </div>\n            ',buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Jobs",text:"\n            <p>Use <strong>Managed Jobs</strong> for long-running workloads.</p>\n            <p>They provide automatic recovery against failures, such as recovering from preemptions or transient errors.</p>\n          ",attachTo:{element:function(){return document.querySelector('a[href="/dashboard/jobs"]')||document.querySelector('a[href="/dashboard/jobs"]')||'a[href="/dashboard/jobs"]'},on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Bring one or many infrastructure",text:"\n            <p>SkyPilot combines your Kubernetes clusters, cloud VMs, or on-premise nodes into a unified compute pool.</p>\n            <p>You can monitor them in this page.</p>\n          ",attachTo:{element:function(){return document.querySelector('a[href="/dashboard/infra"]')||document.querySelector('a[href="/dashboard/infra"]')||'a[href="/dashboard/infra"]'},on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Workspaces",text:"\n            <p>For team deployments, admins can use Workspaces to manage teams or projects.</p>\n          ",attachTo:{element:'a[href="/dashboard/workspaces"]',on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Users",text:"\n            <p>SkyPilot provides user management with RBAC and SSO support. Admins can manage users in this page.</p>\n          ",attachTo:{element:'a[href="/dashboard/users"]',on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"\uD83C\uDF89 Tour complete!",text:'\n              <p>We invite you to to explore the rest of the dashboard.</p>\n              <p>To get started with SkyPilot, refer to <a href="https://docs.skypilot.co/en/latest/getting-started/quickstart.html">Quickstart</a>. You can restart this tour by clicking the <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="inline-block h-4 w-4 align-middle"><circle cx="12" cy="12" r="10"></circle><path d="M9.09 9a3 3 0 0 1 5.83 1c0 2-3 3-3 3"></path><path d="M12 17h.01"></path></svg> icon in the bottom right corner.</p>\n              <p>Have questions? Join Slack to directly chat with the SkyPilot team.</p>\n            ',buttons:[{text:"Finish",action(){this.complete()},classes:"shepherd-button-secondary"},{text:"Join Slack",action(){window.open("https://slack.skypilot.co/","_blank"),this.complete()}}]}].forEach(e=>{n.current.addStep(e)})}return r&&!l&&(g(),u(!0)),()=>{n.current&&n.current.complete()}},[r]),(0,a.useEffect)(()=>{if(!d||f)return;let e=e=>{if(!m.current){if(window.confirm("The tour is currently in progress. Do you want to abort the tour and navigate away?\n\nYou can resume the tour by clicking the question mark on the bottom right."))n.current&&n.current.cancel();else throw o.events.emit("routeChangeError"),"Route change aborted by user during tour."}},t=e=>(e.preventDefault(),e.returnValue="The tour is currently in progress. Are you sure you want to leave?",e.returnValue);return o.events.on("routeChangeStart",e),window.addEventListener("beforeunload",t),()=>{o.events.off("routeChangeStart",e),window.removeEventListener("beforeunload",t)}},[d,f,o]);let y={startTour:g,completeTour:()=>{n.current&&n.current.complete(),p(!1)},tour:n.current};return(0,s.jsx)(t6.Provider,{value:y,children:t})}window.copyDashboardCodeBlock=function(e){let t=e.closest(".bg-gray-50").querySelector("pre").querySelector("code.block").textContent;navigator.clipboard.writeText(t).then(()=>{let t=e.innerHTML;e.innerHTML='<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-check h-5 w-5 text-green-600"><path d="m9 12 2 2 4-4"/></svg>',setTimeout(()=>{e.innerHTML=t},2e3)})}},32586:function(e,t,n){"use strict";n.r(t);var o=n(85893),r=n(67294),i=n(5152),l=n.n(i),s=n(45697),a=n.n(s);n(36501);var c=n(93225),u=n(94849);let d=l()(()=>Promise.all([n.e(616),n.e(6130),n.e(7411),n.e(6989),n.e(3850),n.e(6135),n.e(9666)]).then(n.bind(n,6135)).then(e=>e.Layout),{loadableGenerated:{webpack:()=>[6135]},ssr:!1});function h(e){let{Component:t,pageProps:n}=e;return(0,r.useEffect)(()=>{let e=document.createElement("link");e.rel="icon",e.href="".concat(c.GW,"/favicon.ico"),document.head.appendChild(e)},[]),(0,o.jsx)(u.C,{children:(0,o.jsx)(d,{highlighted:n.highlighted,children:(0,o.jsx)(t,{...n})})})}h.propTypes={Component:a().elementType.isRequired,pageProps:a().object.isRequired},t.default=h},36501:function(){},77663:function(e){!function(){var t={229:function(e){var t,n,o,r=e.exports={};function i(){throw Error("setTimeout has not been defined")}function l(){throw Error("clearTimeout has not been defined")}function s(e){if(t===setTimeout)return setTimeout(e,0);if((t===i||!t)&&setTimeout)return t=setTimeout,setTimeout(e,0);try{return t(e,0)}catch(n){try{return t.call(null,e,0)}catch(n){return t.call(this,e,0)}}}!function(){try{t="function"==typeof setTimeout?setTimeout:i}catch(e){t=i}try{n="function"==typeof clearTimeout?clearTimeout:l}catch(e){n=l}}();var a=[],c=!1,u=-1;function d(){c&&o&&(c=!1,o.length?a=o.concat(a):u=-1,a.length&&h())}function h(){if(!c){var e=s(d);c=!0;for(var t=a.length;t;){for(o=a,a=[];++u<t;)o&&o[u].run();u=-1,t=a.length}o=null,c=!1,function(e){if(n===clearTimeout)return clearTimeout(e);if((n===l||!n)&&clearTimeout)return n=clearTimeout,clearTimeout(e);try{n(e)}catch(t){try{return n.call(null,e)}catch(t){return n.call(this,e)}}}(e)}}function f(e,t){this.fun=e,this.array=t}function p(){}r.nextTick=function(e){var t=Array(arguments.length-1);if(arguments.length>1)for(var n=1;n<arguments.length;n++)t[n-1]=arguments[n];a.push(new f(e,t)),1!==a.length||c||s(h)},f.prototype.run=function(){this.fun.apply(null,this.array)},r.title="browser",r.browser=!0,r.env={},r.argv=[],r.version="",r.versions={},r.on=p,r.addListener=p,r.once=p,r.off=p,r.removeListener=p,r.removeAllListeners=p,r.emit=p,r.prependListener=p,r.prependOnceListener=p,r.listeners=function(e){return[]},r.binding=function(e){throw Error("process.binding is not supported")},r.cwd=function(){return"/"},r.chdir=function(e){throw Error("process.chdir is not supported")},r.umask=function(){return 0}}},n={};function o(e){var r=n[e];if(void 0!==r)return r.exports;var i=n[e]={exports:{}},l=!0;try{t[e](i,i.exports,o),l=!1}finally{l&&delete n[e]}return i.exports}o.ab="//";var r=o(229);e.exports=r}()},5152:function(e,t,n){e.exports=n(19606)},11163:function(e,t,n){e.exports=n(26036)},92703:function(e,t,n){"use strict";var o=n(50414);function r(){}function i(){}i.resetWarningCache=r,e.exports=function(){function e(e,t,n,r,i,l){if(l!==o){var s=Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw s.name="Invariant Violation",s}}function t(){return e}e.isRequired=e;var n={array:e,bigint:e,bool:e,func:e,number:e,object:e,string:e,symbol:e,any:e,arrayOf:t,element:e,elementType:e,instanceOf:t,node:e,objectOf:t,oneOf:t,oneOfType:t,shape:t,exact:t,checkPropTypes:i,resetWarningCache:r};return n.PropTypes=n,n}},45697:function(e,t,n){e.exports=n(92703)()},50414:function(e){"use strict";e.exports="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED"}},function(e){var t=function(t){return e(e.s=t)};e.O(0,[9774,179],function(){return t(91118),t(26036)}),_N_E=e.O()}]);