"""Git related constants."""

GIT_TOKEN_ENV_VAR = 'GIT_TOKEN'
GIT_SSH_KEY_PATH_ENV_VAR = 'GIT_SSH_KEY_PATH'
GIT_SSH_KEY_ENV_VAR = 'GIT_SSH_KEY'
GIT_URL_ENV_VAR = 'GIT_URL'
GIT_COMMIT_HASH_ENV_VAR = 'GIT_COMMIT_HASH'
GIT_BRANCH_ENV_VAR = 'GIT_BRANCH'
GIT_TAG_ENV_VAR = 'GIT_TAG'
